__author__ = "Simon Nilsson"

import os
from copy import deepcopy
from typing import Any, Dict, Optional, Union

import numpy as np
import pandas as pd

pd.options.mode.chained_assignment = None

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from datetime import datetime

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists, check_instance, check_int,
                                check_str, check_that_column_exist,
                                check_valid_boolean)
from simba.utils.data import detect_bouts
from simba.utils.enums import Formats, Options
from simba.utils.errors import DataHeaderError, InvalidInputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (copy_files_to_directory,
                                    find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_df, write_df)

BODY_PART_TYPE = 'body-part'
ANIMAL_TYPE = 'animal'
NEAREST = 'nearest'
LINEAR = 'linear'
QUADRATIC = 'quadratic'

class AdvancedInterpolator(ConfigReader):
    """
    Interpolation method that allows different interpolation parameters for different animals or body-parts.
    For example, interpolate some body-parts of animals using linear interpolation, and other body-parts of animals using nearest interpolation.

    .. image:: _static/img/AdvancedInterpolator.webp
       :width: 600
       :align: center

    .. video:: _static/img/smoothing_example_2.webm
       :width: 800
       :autoplay:
       :loop:

    :param Union[str, os.PathLike] data_path: Path to folder containing pose-estimation data or a file with pose-estimation data.
    :param Union[str, os.PathLike] config_path: Optional path to SimBA project config file in Configparser format.
    :param Literal["animal", "body-part"] type: Type of interpolation: animal or body-part. Default: 'body-part'.
    :param Dict settings: Interpolation rules for each animal or each animal body-part. See examples.
    :param bool verbose: If True, prints progress messages. Default: True.
    :param Union[str, os.PathLike] save_dir: Optional directory to save results. If None, saves in input directory.
    :param bool multi_index_data: If True, the incoming data has multi-index columns. Default: False.
    :param bool save_copy: If True, saves original data in datetime-stamped sub-directory. Default: True.
    :param Optional[int] max_interpolation_length: Maximum length of gaps to interpolate. If None, interpolates all gaps. Default: None.

    :examples:
    >>> # Animal-level interpolation
    >>> interpolator = AdvancedInterpolator(
    ...     data_path='/path/to/project_folder/csv/input_csv',
    ...     config_path='/path/to/project_folder/project_config.ini',
    ...     type='animal',
    ...     settings={'Animal_1': 'linear', 'Animal_2': 'quadratic'},
    ...     multi_index_data=True
    ... )
    >>> interpolator.run()
    >>> 
    >>> # Body-part level interpolation
    >>> interpolator = AdvancedInterpolator(
    ...     data_path='/path/to/project_folder/csv/input_csv',
    ...     config_path='/path/to/project_folder/project_config.ini',
    ...     type='body-part',
    ...     settings={
    ...         'Simon': {
    ...             'Ear_left_1': 'linear',
    ...             'Ear_right_1': 'linear',
    ...             'Nose_1': 'quadratic',
    ...             'Lat_left_1': 'quadratic',
    ...             'Lat_right_1': 'quadratic',
    ...             'Center_1': 'nearest',
    ...             'Tail_base_1': 'nearest'
    ...         },
    ...         'JJ': {
    ...             'Ear_left_2': 'nearest',
    ...             'Ear_right_2': 'nearest',
    ...             'Nose_2': 'quadratic',
    ...             'Lat_left_2': 'quadratic',
    ...             'Lat_right_2': 'quadratic',
    ...             'Center_2': 'linear',
    ...             'Tail_base_2': 'linear'
    ...         }
    ...     },
    ...     multi_index_data=True
    ... )
    >>> interpolator.run()
    """

    def __init__(self,
                 data_path: Union[str, os.PathLike],
                 settings: Dict[str, Any],
                 type: Optional[Literal["animal", "body-part"]] = 'body-part',
                 verbose: Optional[bool] = True,
                 config_path: Optional[Union[str, os.PathLike]] = None,
                 save_dir: Optional[Union[str, os.PathLike]] = None,
                 multi_index_data: Optional[bool] = False,
                 save_copy: Optional[bool] = True,
                 max_interpolation_length: Optional[int] = None):



        if config_path is not None:
            ConfigReader.__init__(self, config_path=config_path, read_video_info=False, create_logger=False)
        else:
            self.file_type, self.datetime = Formats.CSV.value, datetime.now().strftime("%Y%m%d%H%M%S")
            self.timer = SimbaTimer(start=True)
        check_str(name=f'{self.__class__.__name__} type', value=type, options=["animal", "body-part"], raise_error=True)
        if os.path.isfile(data_path):
            check_file_exist_and_readable(file_path=data_path)
            self.file_paths = [data_path]
            self.input_dir = os.path.dirname(data_path)
            self.cpy_dir = os.path.join(os.path.dirname(data_path), f"Pre_Advanced_Interpolation_{self.datetime}")
        elif os.path.isdir(data_path):
            self.file_paths = find_files_of_filetypes_in_directory(directory=data_path, extensions=[f".{self.file_type}"], raise_warning=False,raise_error=True)
            self.cpy_dir = os.path.join(data_path, f"Pre_Advanced_Interpolation_{self.datetime}")
            self.input_dir = data_path
        else:
            raise InvalidInputError(msg=f'{data_path} is not a valid file path or file directory', source=self.__class__.__name__)
        if save_dir is not None:
            check_if_dir_exists(in_dir=save_dir)
        else:
            save_dir = self.input_dir

        check_valid_boolean(value=[multi_index_data, save_copy], source=self.__class__.__name__, raise_error=True)
        check_instance(source=self.__class__.__name__, instance=settings, accepted_types=(dict,))
        for animal, animal_data in settings.items():
            if type == BODY_PART_TYPE:
                check_instance(source=self.__class__.__name__, instance=animal_data, accepted_types=(dict,))
                for bp_name, bp_data in animal_data.items():
                    if config_path is not None:
                        check_str(name='method', value=bp_name, options=self.project_bps)
                    check_str(name='method', value=bp_data, options=[LINEAR, NEAREST, QUADRATIC])
            else:
               check_str(name='method', value=animal_data, options=[LINEAR, NEAREST, QUADRATIC])
        self.settings, self.type, self.multi_index_data, self.verbose = settings, type, multi_index_data, verbose
        if type == ANIMAL_TYPE and config_path is not None: self.__transpose_settings()
        self.save_copy, self.save_dir, self.config_path = save_copy, save_dir, config_path
        if max_interpolation_length is not None:
            check_int(name=f'{self.__class__.__name__} max_interpolation_length', min_value=1, raise_error=True, value=max_interpolation_length)
        self.max_interpolation_length = max_interpolation_length
        if save_copy and not os.path.isdir(self.cpy_dir): os.makedirs(self.cpy_dir)

    def __transpose_settings(self):
        """Helper to transpose settings dict if interpolating per animal, so the same method can be used for both animal and body-part interpolation"""
        transposed_settings = {}
        for animal_name, body_part_data in self.animal_bp_dict.items():
            transposed_settings[animal_name] = {}
            for animal_body_part in body_part_data["X_bps"]:
                transposed_settings[animal_name][animal_body_part[:-2]] = self.settings[animal_name]
        self.settings = transposed_settings

    def __insert_multi_index(self, df: pd.DataFrame) -> pd.DataFrame:
        multi_idx_header = []
        for i in range(len(df.columns)):
            multi_idx_header.append(("IMPORTED_POSE", "IMPORTED_POSE", list(df.columns)[i]))
        df.columns = pd.MultiIndex.from_tuples(multi_idx_header)
        return df

    def run(self):
        for file_cnt, file_path in enumerate(self.file_paths):
            video_timer = SimbaTimer(start=True)
            _, video_name, _ = get_fn_ext(filepath=file_path)
            if self.config_path is not None:
                df = read_df(file_path=file_path, file_type=self.file_type, check_multiindex=self.multi_index_data).fillna(0).reset_index(drop=True)
                if len(df.columns) != len(self.bp_col_names):
                    raise DataHeaderError(msg=f"The SimBA project suggest the data should have {len(self.bp_col_names)} columns, but the input data has {len(df.columns)} columns", source=self.__class__.__name__)
                df.columns = self.bp_headers
                df[df < 0] = 0
            else:
                df = pd.read_csv(filepath_or_buffer=file_path, index_col=0)

            df.columns = [x.lower() for x in df.columns]
            df_cpy = deepcopy(df)
            for animal_name, animal_body_parts in self.settings.items():
                for bp, interpolation_setting in animal_body_parts.items():
                    bp = bp.lower()
                    check_that_column_exist(df=df, column_name=[f"{bp}_x", f"{bp}_y"], file_name=file_path)
                    df[[f"{bp}_x", f"{bp}_y"]] = df[[f"{bp}_x", f"{bp}_y"]].astype(np.int32)
                    if self.max_interpolation_length is None:
                        df[df <= 0] = 0
                        idx = df.loc[(df[f"{bp}_x"] <= 0.0) & (df[f"{bp}_y"] <= 0.0)].index.tolist()
                        if self.verbose: print(f"Interpolating {len(idx)} {bp} body-parts in video {video_name}...")
                        df[[f"{bp}_x", f"{bp}_y"]] = (df[[f"{bp}_x", f"{bp}_y"]].interpolate(method=interpolation_setting, axis=0).ffill().bfill().astype(np.int32))
                        df[[f"{bp}_x", f"{bp}_y"]][df[[f"{bp}_x", f"{bp}_y"]] < 0] = 0
                    else:
                        df_cpy.loc[df_cpy[f'{bp}_x'] <= 0, f'{bp}_x'] = 0
                        df_cpy.loc[df_cpy[f'{bp}_y'] <= 0, f'{bp}_y'] = 0
                        idx = df.loc[(df[f"{bp}_x"] <= 0.0) & (df[f"{bp}_y"] <= 0.0)].index.tolist()
                        df_cpy[f'{bp}_temp'] = 0
                        df_cpy.loc[idx, [f'{bp}_temp']] = 1
                        bouts = detect_bouts(data_df=df_cpy, target_lst=[f'{bp}_temp'], fps=1)
                        bouts = bouts[bouts['Bout_time'] <= self.max_interpolation_length]
                        if len(bouts) > 0:
                            idx = bouts.apply(lambda row: list(range(row['Start_frame'], row['End_frame'] + 1)), axis=1).explode().tolist()
                            if self.verbose: print(f"Interpolating {len(idx)} {bp} body-parts in video {video_name}...")
                            df.loc[idx, [f"{bp}_x", f"{bp}_y"]] = np.nan
                            df[[f"{bp}_x", f"{bp}_y"]] = (df[[f"{bp}_x", f"{bp}_y"]].interpolate(method=interpolation_setting, axis=0).astype(np.int32))
            if self.multi_index_data:
                df = self.__insert_multi_index(df=df)
            if self.save_copy:
                copy_files_to_directory(file_paths=[file_path], dir=self.cpy_dir, verbose=False)
            save_path = os.path.join(self.save_dir, f'{video_name}{self.file_type}')
            write_df(df=df, file_type=self.file_type, save_path=save_path, multi_idx_header=self.multi_index_data)
            video_timer.stop_timer()
            print(f'Video {video_name} complete. Elapsed time {video_timer.elapsed_time_str}s')
        self.timer.stop_timer()
        if self.save_copy:
            msg = f"Advanced interpolation complete. Data saved in {self.save_dir}. Original copies saved in {self.cpy_dir}."
        else:
            msg = f"Advanced interpolation complete. Data saved in {self.save_dir}."
        stdout_success(msg=msg, elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)

# SMOOTHING_SETTINGS = {'Simon': {'Ear_left_1': {'method': 'Savitzky Golay', 'time_window': 3500},
#                                'Ear_right_1': {'method': 'Gaussian', 'time_window': 500},
#                                'Nose_1': {'method': 'Savitzky Golay', 'time_window': 2000},
#                                'Lat_left_1': {'method': 'Savitzky Golay', 'time_window': 2000},
#                                'Lat_right_1': {'method': 'Gaussian', 'time_window': 2000},
#                                'Center_1': {'method': 'Savitzky Golay', 'time_window': 2000},
#                                'Tail_base_1': {'method': 'Gaussian', 'time_window': 500}},
#                         'JJ': {'Ear_left_2': {'method': 'Savitzky Golay', 'time_window': 2000},
#                                'Ear_right_2': {'method': 'Savitzky Golay', 'time_window': 500},
#                                'Nose_2': {'method': 'Gaussian', 'time_window': 3500},
#                                'Lat_left_2': {'method': 'Savitzky Golay', 'time_window': 500},
#                                'Lat_right_2': {'method': 'Gaussian', 'time_window': 3500},
#                                'Center_2': {'method': 'Gaussian', 'time_window': 2000},
#                                'Tail_base_2': {'method': 'Savitzky Golay', 'time_window': 3500}}}
#
#
# INTERPOLATION_SETTINGS = {'Animal_1': {'NOSE': 'linear',
#                           'LEFT_EAR': 'linear',
#                           'RIGHT_EAR': 'quadratic',
#                           'LEFT_SIDE': 'quadratic',
#                           'CENTER': 'quadratic',
#                           'RIGHT_SIDE': 'nearest',
#                           'TAIL_BASE': 'nearest'}}

# advanced_interpolator = AdvancedInterpolator(data_path=r'D:\netholabs\data', settings=INTERPOLATION_SETTINGS, type='body-part', multi_index_data=True, save_copy=False, max_interpolation_length=100)

# advanced_interpolator.run()

# for animal, animal_data in settings.items():
#     check_instance(source=self.__class__.__name__, instance=animal_data, accepted_types=(dict,))
#     if type == BODY_PART_TYPE:
#         for bp_name, bp_data in animal_data.items():
#             check_if_keys_exist_in_dict(data=bp_data, key=['method', 'time_window'])
#             check_str(name='method', value=bp_data['method'], options=[GAUSSIAN, SAVITZKY_GOLAY])
#             check_int(name='time_window', value=bp_data['time_window'], min_value=1)
#     else:
#         check_if_keys_exist_in_dict(data=animal_data, key=['method', 'time_window'])
#         check_str(name='method', value=animal_data['method'], options=[GAUSSIAN, SAVITZKY_GOLAY])
#         check_int(name='time_window', value=animal_data['time_window'], min_value=1)


__author__ = "Simon Nilsson"

import datetime
import itertools
import os
import random
from copy import deepcopy
from typing import Any, Dict, Optional, Union

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import pandas as pd

from simba.mixins.train_model_mixin import TrainModelMixin
from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.unsupervised.enums import Unsupervised
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists,
                                check_if_keys_exist_in_dict,
                                check_if_list_contains_values, check_instance,
                                check_str, check_that_directory_is_empty,
                                check_umap_hyperparameters)
from simba.utils.enums import Formats
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (drop_df_fields, read_pickle, write_df,
                                    write_pickle)

try:
    from cuml import UMAP

    gpu_flag = True
except ModuleNotFoundError:
    from umap import UMAP


class UmapEmbedder(UnsupervisedMixin):
    """
    Methods for grid-search UMAP model fit and transform.
    Defaults to GPU and cuml.UMAP if GPU available. If GPU unavailable, then umap.UMAP.

    :param data_path: Path holding pickled data-set created by `simba.unsupervised.dataset_creator.DatasetCreator.
    :param save_dir: Empty directory where to save the UMAP results.
    :param hyper_parameters: dict holding UMAP hyperparameters in list format.

    :Example I: Fit.
    >>> hyper_parameters = {'n_neighbors': [10, 2], 'min_distance': [1.0], 'spread': [1.0], 'scaler': 'MIN-MAX', 'variance': 0.25, "multicolinearity": 0.5}
    >>> data_path = 'unsupervised/project_folder/logs/unsupervised_data_20230416145821.pickle'
    >>> save_dir = 'unsupervised/dr_models'
    >>> config_path = 'unsupervised/project_folder/project_config.ini'
    >>> embedder = UmapEmbedder(data_path=data_path, save_dir=save_dir)
    >>> embedder.fit(hyper_parameters=hyper_parameters)
    """

    def __init__(self):
        self.datetime = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        super().__init__()

    def fit(
        self,
        data_path: Union[str, os.PathLike],
        save_dir: [str, os.PathLike],
        hyper_parameters: dict,
    ):
        check_file_exist_and_readable(file_path=data_path)
        check_that_directory_is_empty(directory=save_dir)
        self.data_path = data_path
        self.data = read_pickle(data_path=data_path)
        self.umap_df = (
            deepcopy(self.data[Unsupervised.BOUTS_FEATURES.value])
            .reset_index()
            .set_index(
                [
                    Unsupervised.VIDEO.value,
                    Unsupervised.START_FRAME.value,
                    Unsupervised.END_FRAME.value,
                ]
            )
        )
        self.save_dir = save_dir

        self.low_var_cols, self.hyper_parameters = None, hyper_parameters
        check_umap_hyperparameters(hyper_parameters=hyper_parameters)
        self.search_space = list(
            itertools.product(
                *[
                    hyper_parameters[Unsupervised.N_NEIGHBORS.value],
                    hyper_parameters[Unsupervised.MIN_DISTANCE.value],
                    hyper_parameters[Unsupervised.SPREAD.value],
                ]
            )
        )
        print(f"Building {len(self.search_space)} UMAP model(s)...")
        if hyper_parameters[Unsupervised.VARIANCE.value] > 0:
            self.low_var_cols = TrainModelMixin.find_low_variance_fields(
                data=self.umap_df,
                variance_threshold=hyper_parameters[Unsupervised.VARIANCE.value],
            )
            self.umap_df = drop_df_fields(data=self.umap_df, fields=self.low_var_cols)
        # if hyper_parameters[Unsupervised.MULTICOLLINEARITY.value] > 0:
        #     print(hyper_parameters[Unsupervised.MULTICOLLINEARITY.value])
        self.scaler = TrainModelMixin.define_scaler(
            scaler_name=hyper_parameters[Unsupervised.SCALER.value]
        )
        self.scaler.fit(self.umap_df)
        self.scaled_umap_data = TrainModelMixin.scaler_transform(
            data=self.umap_df, scaler=self.scaler
        )
        self.__create_methods_log()
        self.__fit_umaps()
        self.timer.stop_timer()
        stdout_success(
            msg=f"{len(self.search_space)} models saved in {self.save_dir} directory",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def __create_methods_log(self):
        self.methods = {}
        self.methods[Unsupervised.SCALER.value] = self.scaler
        self.methods[Unsupervised.SCALER_TYPE.value] = self.hyper_parameters[
            Unsupervised.SCALER.value
        ]
        self.methods[Unsupervised.SCALED_DATA.value] = self.scaled_umap_data
        self.methods[Unsupervised.VARIANCE.value] = self.hyper_parameters[
            Unsupervised.VARIANCE.value
        ]
        self.methods[Unsupervised.LOW_VARIANCE_FIELDS.value] = self.low_var_cols
        self.methods[Unsupervised.FEATURE_NAMES.value] = self.scaled_umap_data.columns

    def __fit_umaps(self):
        for cnt, h in enumerate(self.search_space):
            self.model_count = cnt
            self.model = {}
            self.model_timer = SimbaTimer()
            self.model_timer.start_timer()
            self.model[Unsupervised.HASHED_NAME.value] = random.sample(
                self.model_names, 1
            )[0]
            self.model[Unsupervised.PARAMETERS.value] = {
                Unsupervised.N_NEIGHBORS.value: h[0],
                Unsupervised.MIN_DISTANCE.value: h[1],
                Unsupervised.SPREAD.value: h[2],
            }
            self.model[Unsupervised.MODEL.value] = UMAP(
                min_dist=self.model[Unsupervised.PARAMETERS.value][
                    Unsupervised.MIN_DISTANCE.value
                ],
                n_neighbors=int(
                    self.model[Unsupervised.PARAMETERS.value][
                        Unsupervised.N_NEIGHBORS.value
                    ]
                ),
                spread=self.model[Unsupervised.PARAMETERS.value][
                    Unsupervised.SPREAD.value
                ],
                metric=Unsupervised.EUCLIDEAN.value,
                verbose=0,
            )
            self.model[Unsupervised.MODEL.value].fit(self.scaled_umap_data.values)
            results = {}
            results[Unsupervised.DATA.value] = self.data
            results[Unsupervised.METHODS.value] = self.methods
            results[Unsupervised.DR_MODEL.value] = self.model
            self.__save(data=results)

    def __save(self, data: dict) -> None:
        write_pickle(
            data=data,
            save_path=os.path.join(
                self.save_dir,
                f"{self.model[Unsupervised.HASHED_NAME.value]}.{Formats.PICKLE.value}",
            ),
        )
        self.model_timer.stop_timer()
        stdout_success(
            msg=f"Model {self.model_count+1}/{len(self.search_space)} ({self.model[Unsupervised.HASHED_NAME.value]}) saved...",
            elapsed_time=self.model_timer.elapsed_time,
        )

    def transform(
        self,
        data_path: Union[str, os.PathLike],
        model: Union[str, os.PathLike, dict],
        settings: Optional[Dict[str, Any]] = None,
        save_dir: Optional[os.PathLike] = None,
        save_format: Optional[Literal["csv", "pickle"]] = "csv",
    ) -> Union[None, pd.DataFrame]:
        """
        :Example I: Transform.
        >>> data_path = '/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/logs/unsupervised_data_20240215143716.pickle'
        >>> model_path = '/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/clustering1704/academic_montalcini.pickle'
        >>> embedder = UmapEmbedder()
        >>> embedder.transform(save_dir=None, data_path=data_path, model=model_path, settings=None)
        >>> embedder.transform(save_dir=None, data_path=data_path, model=model_path, settings={'DATA_FORMAT': 'scaled', 'CLASSIFICATIONS': True})
        """
        print("Transforming data using UMAP model...")
        timer = SimbaTimer(start=True)
        check_file_exist_and_readable(file_path=data_path)
        if save_dir is not None:
            check_if_dir_exists(in_dir=save_dir)
            check_str(
                name=f"{UmapEmbedder.transform.__name__} save_format",
                value=save_format,
                options=(
                    "csv",
                    "pickle",
                ),
            )
        if isinstance(model, str):
            check_file_exist_and_readable(file_path=model)
            model = read_pickle(data_path=model)
        data = read_pickle(data_path=data_path)
        check_if_keys_exist_in_dict(
            data=data,
            key=[Unsupervised.BOUTS_FEATURES.value, Unsupervised.BOUTS_TARGETS.value],
            name=data_path,
        )
        check_if_keys_exist_in_dict(
            data=model, key=[Unsupervised.METHODS.value], name=data_path
        )
        mdl_name = model[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]
        umap_df = deepcopy(data[Unsupervised.BOUTS_FEATURES.value])
        umap_df = self.drop_fields(
            data=umap_df,
            fields=model[Unsupervised.METHODS.value][
                Unsupervised.LOW_VARIANCE_FIELDS.value
            ],
        )

        scaled_umap_data = TrainModelMixin.scaler_transform(
            data=umap_df,
            scaler=model[Unsupervised.METHODS.value][Unsupervised.SCALER.value],
        )
        scaled_umap_data = drop_df_fields(
            data=scaled_umap_data,
            fields=model[Unsupervised.METHODS.value][
                Unsupervised.LOW_VARIANCE_FIELDS.value
            ],
        )
        check_if_list_contains_values(
            data=list(scaled_umap_data.columns),
            values=model[Unsupervised.METHODS.value][Unsupervised.FEATURE_NAMES.value],
            name=self.__class__.__name__,
        )
        results = pd.DataFrame(
            model[Unsupervised.DR_MODEL.value][Unsupervised.MODEL.value].transform(
                scaled_umap_data
            ),
            columns=["X", "Y"],
            index=umap_df.index,
        )
        if settings is not None:
            check_instance(
                source=f"{UmapEmbedder.transform.__name__} settings",
                instance=settings,
                accepted_types=(dict,),
            )
            if "DATA_FORMAT" in settings.keys():
                if settings["DATA_FORMAT"] == Unsupervised.SCALED.value:
                    results = pd.concat([scaled_umap_data, results], axis=1)
                elif settings["DATA_FORMAT"] == Unsupervised.RAW.value:
                    results = pd.concat([umap_df, results], axis=1)
            if "CLASSIFICATIONS" in settings.keys():
                target_data = data[Unsupervised.BOUTS_TARGETS.value].set_index(
                    ["VIDEO", "START_FRAME", "END_FRAME"]
                )
                results = results.join(target_data, how="inner")
        if save_dir:
            if save_format == Formats.CSV.value:
                save_path = os.path.join(
                    save_dir,
                    f"Transformed_{mdl_name}_{self.datetime}.{Formats.CSV.value}",
                )
                results.to_csv(path_or_buf=save_path)
            else:
                save_path = os.path.join(
                    save_dir,
                    f"Transformed_{mdl_name}_{self.datetime}.{Formats.PICKLE.value}",
                )
                write_pickle(data=results, save_path=save_path)
            timer.stop_timer()
            stdout_success(
                msg=f"Transformed data saved at {save_path} (elapsed time: {timer.elapsed_time_str}s)"
            )
        else:
            timer.stop_timer()
            stdout_success(
                msg=f"Data transformed using UMAP model {mdl_name} (elapsed time: {timer.elapsed_time_str}s)"
            )
            return results


hyper_parameters = {
    "n_neighbors": [5],
    "min_distance": [0.1, 0.5, 0.0],
    "spread": [1.0],
    "scaler": "MIN-MAX",
    "variance": 0.25,
    "multicollinearity": 0.7,
}
data_path = "/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/logs/unsupervised_data_20240325092459.pickle"
save_dir = "/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/embedding_2"
config_path = (
    "/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini"
)
embedder = UmapEmbedder()
embedder.fit(data_path=data_path, save_dir=save_dir, hyper_parameters=hyper_parameters)
#

# data_path = '/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/logs/unsupervised_data_20240215143716.pickle'
# model_path = '/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/clustering1704/academic_montalcini.pickle'
# embedder = UmapEmbedder()
# #embedder.transform(save_dir=None, data_path=data_path, model=model_path, settings=None)
#
# embedder.transform(save_dir='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/new_viz', data_path=data_path, model=model_path, settings={'DATA_FORMAT': 'scaled', 'CLASSIFICATIONS': True})


# data_path = '/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/logs/unsupervised_data_20240215143716.pickle'
# model_path = '/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/clustering1704/academic_montalcini.pickle'
# embedder = UmapEmbedder()
# #embedder.transform(save_dir=None, data_path=data_path, model=model_path, settings=None)
#
# embedder.transform(save_dir='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/new_viz', data_path=data_path, model=model_path, settings={'DATA_FORMAT': 'scaled', 'CLASSIFICATIONS': True})

#

# data_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/logs/unsupervised_data_20230416145821.pickle'
# save_dir = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/transformed_umap'
# settings = {'DATA': 'RAW', 'format': 'csv'}
# embedder = UmapEmbedder(data_path=data_path, save_dir=save_dir)
# embedder.transform(model='/Users/simon/Desktop/envs/troubleshooting/unsupervised/dr_models/boring_lederberg.pickle', settings=settings)


# model_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/dr_models/funny_heisenberg.pickle'
# data_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/logs/unsupervised_data_20230222150701.pickle'
# save_dir = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/dr_models/'
# _ = UMAPTransform(model_path=model_path, data_path=data_path, save_dir=save_dir, settings=settings)

#
#
#
#
#
#
# hyper_parameters = {'n_neighbors': [10, 2], 'min_distance': [1.0], 'spread': [1.0], 'scaler': 'MIN-MAX', 'variance': 0.25}
# data_path = '/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/logs/unsupervised_data_20240214093117.pickle'
# save_dir = '/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/dim_reduction_mdls'
# config_path = '/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/project_config.ini'
# embedder = UmapEmbedder()
# embedder.fit(data_path=data_path, save_dir=save_dir, hyper_parameters=hyper_parameters)



import os
import random
from typing import Optional, Tuple, Union

import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import cv2
import numpy as np

from simba.mixins.pose_importer_mixin import PoseImporterMixin
from simba.third_party_label_appenders.transform.utils import (
    create_yolo_keypoint_yaml, get_yolo_keypoint_flip_idx)
from simba.utils.checks import (check_float, check_if_dir_exists, check_int,
                                check_str, check_valid_boolean,
                                check_valid_tuple)
from simba.utils.enums import Formats, Options
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (create_directory,
                                    find_files_of_filetypes_in_directory,
                                    get_video_meta_data, read_frm_of_video)
from simba.utils.warnings import FrameRangeWarning
from simba.utils.yolo import keypoint_array_to_yolo_annotation_str

H5_EXT = ['.H5', '.h5']

class MADLCH52Yolo:
    """
    Convert multi-animal DeepLabCut pose estimation H5 data and corresponding videos into YOLO keypoint dataset format.

    .. note::
       This converts DeepLabCut **inference** data to YOLO keypoints (not DeepLabcut annotations).

    :param Union[str, os.PathLike] data_dir: Directory path containing DLC-generated H5 files with inferred keypoints.
    :param Union[str, os.PathLike] video_dir: Directory path containing corresponding videos from which frames are to be extracted.
    :param Union[str, os.PathLike] save_dir: Output directory where YOLO-formatted images, labels, and map YAML file will be saved. Subdirectories `images/train`, `images/val`, `labels/train`, `labels/val` will be created.
    :param Optional[int] frms_cnt: Number of frames to randomly sample from each video for conversion. If None, all frames are used.
    :param float threshold: Minimum confidence score threshold to filter out low-confidence pose instances. Only instances with `instance.score` >= this threshold are used.
    :param float train_size: Proportion of frames randomly assigned to the training dataset. Value must be between 0.1 and 0.99. Default: 0.7.
    :param bool verbose: If True, prints progress. Default: True.
    :param Tuple[int, ...] flip_idx: Tuple of keypoint indices used for horizontal flip augmentation during training. The tuple defines the order of keypoints after flipping. If None, it will be inferred.
    :param float padding: Fractional padding to add around the bounding boxes (relative to image dimensions). Helps to slightly enlarge bounding boxes by this percentage. Default 0.05. E.g., Useful when all body-parts are along animal length.
    :param Optional[str] single_id: If the data contains pose-estimation for multiple indivisuals, but you want to treat it as examples of a single individual, pass the name of the single individual. Defaults to None, and the YOLO data will be formatted to the number of objects which the H5 data contains.
    :return: None. Results saved in ``save_dir``.

    :example:
    >>> DATA_DIR = r'D:\troubleshooting\dlc_h5_multianimal_to_yolo\data'
    >>> VIDEO_DIR = r'D:\troubleshooting\dlc_h5_multianimal_to_yolo\videos'
    >>> SAVE_DIR = r"D:\imgs\madlc"
    >>> runner = MADLCH52Yolo(data_dir=DATA_DIR, video_dir=VIDEO_DIR, save_dir=SAVE_DIR, clahe=True, single_id='animal_1')
    >>> runner.run()
    """

    def __init__(self,
                 data_dir: Union[str, os.PathLike],
                 video_dir: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 frms_cnt: Optional[Union[int, None]] = None,
                 verbose: bool = True,
                 threshold: float = 0,
                 train_size: float = 0.7,
                 flip_idx: Tuple[int, ...] = None,
                 greyscale: bool = False,
                 clahe: bool = False,
                 padding: float = 0.00,
                 single_id: Optional[str] = None):


        data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=H5_EXT, as_dict=True, raise_error=True)
        video_paths = find_files_of_filetypes_in_directory(directory=video_dir, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value, as_dict=True, raise_error=True)
        check_if_dir_exists(in_dir=save_dir)
        self.img_dir, self.lbl_dir = os.path.join(save_dir, 'images'), os.path.join(save_dir, 'labels')
        self.img_train_dir, self.img_val_dir = os.path.join(save_dir, 'images', 'train'), os.path.join(save_dir, 'images', 'val')
        self.lbl_train_dir, self.lb_val_dir = os.path.join(save_dir, 'labels', 'train'), os.path.join(save_dir, 'labels', 'val')
        if flip_idx is not None: check_valid_tuple(x=flip_idx, source=f'{self.__class__.__name__} flip_idx', valid_dtypes=(int,), minimum_length=1)
        if single_id is not None: check_str(name=f'{self.__class__.__name__} single_id', value=single_id, raise_error=True)
        check_float(name=f'{self.__class__.__name__} threshold', min_value=0.0, max_value=1.0, raise_error=True, value=threshold)
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose', raise_error=True)
        check_valid_boolean(value=greyscale, source=f'{self.__class__.__name__} greyscale', raise_error=True)
        check_valid_boolean(value=clahe, source=f'{self.__class__.__name__} clahe', raise_error=True)
        check_float(name=f'{self.__class__.__name__} train_size', value=train_size, max_value=0.99, min_value=0.1)
        check_float(name=f'{self.__class__.__name__} padding', value=padding, max_value=1.0, min_value=0.0, raise_error=True)
        self.data_and_videos_lk = PoseImporterMixin().link_video_paths_to_data_paths(data_paths=list(data_paths.values()), video_paths=list(video_paths.values()), str_splits=Formats.DLC_NETWORK_FILE_NAMES.value)
        self.map_path = os.path.join(save_dir, 'map.yaml')
        if frms_cnt is not None: check_int(name=f'{self.__class__ .__name__} frms_cnt', value=frms_cnt, min_value=1, raise_error=True)
        create_directory(paths=[self.img_train_dir, self.img_val_dir, self.lbl_train_dir, self.lb_val_dir])
        self.threshold, self.frms_cnt, self.train_size, self.save_dir = threshold, frms_cnt, train_size, save_dir
        self.verbose, self.greyscale, self.flip_idx, self.padding = verbose, greyscale, flip_idx, padding
        self.clahe, self.single_id = clahe, single_id


    def run(self):
        dfs, timer, bp_header = [], SimbaTimer(start=True), []
        for cnt, (video_name, video_data) in enumerate(self.data_and_videos_lk.items()):
            data = pd.read_hdf(video_data["DATA"], header=[0, 1, 2, 3]).replace([np.inf, -np.inf], np.nan).fillna(-1)
            new_cols = [(x[1], x[2], x[3]) for x in list(data.columns)]
            if cnt == 0:
                self.animal_names = list(dict.fromkeys([x[1] for x in list(data.columns)]))
            data.columns = new_cols
            p_data = data[[x for x in data.columns if 'likelihood' in x]]
            data = data.iloc[(p_data[(p_data > self.threshold).all(axis=1)].index)]
            bp_header = [f'{x[1]}_{x[2]}' for x in data.columns]
            bp_header = [x for x in bp_header if 'likelihood' not in x]
            data['frm_idx'] = data.index
            data['video'] = video_data['VIDEO']
            selected_frms = random.sample(list(data['frm_idx'].unique()), self.frms_cnt) if self.frms_cnt is not None else list(data['frm_idx'].unique())
            data = data[data['frm_idx'].isin(selected_frms)]
            dfs.append(data)

        dfs = pd.concat(dfs, axis=0)
        dfs['id'] = dfs['frm_idx'].astype(str) + dfs['video'].astype(str)
        train_idx = random.sample(list(dfs['id'].unique()), int(len(dfs['frm_idx'].unique()) * self.train_size))
        map_dict = {k: v for k, v in enumerate(self.animal_names)} if self.single_id is None else {0: self.single_id}
        id_dict = {v: k for k, v in enumerate(self.animal_names)}
        if self.flip_idx is None:
            self.flip_idx = get_yolo_keypoint_flip_idx(x=list(dict.fromkeys([x[:-2] for x in bp_header])))

        for frm_cnt, frm_id in enumerate(dfs['id'].unique()):
            frm_data = dfs[dfs['id'] == frm_id]
            video_path = list(frm_data['video'])[0]
            frm_idx = list(frm_data['frm_idx'])[0]
            video_meta = get_video_meta_data(video_path=video_path)
            if self.verbose:
                print(f'Processing frame: {frm_cnt + 1}/{len(dfs)} ...')
            if frm_idx > video_meta['frame_count']:
                FrameRangeWarning(msg=f'Frame {frm_idx} could not be read from video {video_path}. The video {video_meta["video_name"]} has {video_meta["frame_count"]} frames', source=self.__class__.__name__)
                continue
            img = read_frm_of_video(video_path=video_path, frame_index=frm_idx, greyscale=self.greyscale, clahe=self.clahe)
            img_h, img_w = img.shape[0], img.shape[1]
            if list(frm_data['id'])[0] in train_idx:
                img_save_path = os.path.join(self.img_train_dir, f'{video_meta["video_name"]}_{frm_idx}.png')
                lbl_save_path = os.path.join(self.lbl_train_dir, f'{video_meta["video_name"]}_{frm_idx}.txt')
            else:
                img_save_path = os.path.join(self.img_val_dir, f'{video_meta["video_name"]}_{frm_idx}.png')
                lbl_save_path = os.path.join(self.lb_val_dir, f'{video_meta["video_name"]}_{frm_idx}.txt')
            img_lbl = ''
            frm_data = frm_data.drop(['video', 'id', 'frm_idx'], axis=1).T.iloc[:, 0]
            for track_cnt, animal_name in enumerate(self.animal_names):
                keypoints = frm_data.T[[idx[0] == animal_name for idx in frm_data.index]].values.reshape(-1, 3)
                keypoints[keypoints[:, 2] != 0.0, 2] = 2
                instance_str = f'{id_dict[animal_name]} ' if self.single_id is None else f'0 '
                instance_str += keypoint_array_to_yolo_annotation_str(x=keypoints, img_h=img_h, img_w=img_w, padding=self.padding)
                img_lbl += instance_str
            with open(lbl_save_path, mode='wt', encoding='utf-8') as f:
                f.write(img_lbl)
            cv2.imwrite(img_save_path, img)
        create_yolo_keypoint_yaml(path=self.save_dir, train_path=self.img_train_dir, val_path=self.img_val_dir, names=map_dict, save_path=self.map_path, kpt_shape=(len(self.flip_idx), 3), flip_idx=self.flip_idx)
        timer.stop_timer()
        stdout_success(msg=f'YOLO formated data saved in {self.save_dir} directory', source=self.__class__.__name__, elapsed_time=timer.elapsed_time_str)


# DATA_DIR = r'D:\troubleshooting\dlc_h5_multianimal_to_yolo\data'
# VIDEO_DIR = r'D:\troubleshooting\dlc_h5_multianimal_to_yolo\videos'
# SAVE_DIR = r"D:\imgs\madlc"
# runner = MADLCH52Yolo(data_dir=DATA_DIR, video_dir=VIDEO_DIR, save_dir=SAVE_DIR, clahe=True, single_id='animal_1')
# runner.run()


import os
from typing import List, Literal, Optional, Tuple, Union

os.environ["KMP_DUPLICATE_LIB_OK"]="TRUE"
import numpy as np
import pandas as pd
import torch

from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_int, check_valid_boolean,
                                check_valid_lst, check_valid_tuple, get_fn_ext)
from simba.utils.errors import CountError, InvalidFileTypeError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_video_meta_data
from simba.utils.yolo import (_get_undetected_obs, filter_yolo_keypoint_data,
                              load_yolo_model)

OUT_COLS = ['FRAME', 'CLASS_ID', 'CLASS_NAME', 'CONFIDENCE', 'TRACK', 'X1', 'Y1', 'X2', 'Y2', 'X3', 'Y3', 'X4', 'Y4']
COORD_COLS = ['X1', 'Y1', 'X2', 'Y2', 'X3', 'Y3', 'X4', 'Y4']
NEAREST = 'nearest'

class YOLOPoseTrackInference():


    def __init__(self,
                 weights_path: Union[str, os.PathLike],
                 video_path: Union[Union[str, os.PathLike], List[Union[str, os.PathLike]]],
                 keypoint_names: Tuple[str, ...],
                 config_path: Union[str, os.PathLike],
                 verbose: Optional[bool] = False,
                 save_dir: Optional[Union[str, os.PathLike]] = None,
                 device: Union[Literal['cpu'], int] = 0,
                 format: Optional[str] = None,
                 batch_size: Optional[int] = 4,
                 torch_threads: int = 8,
                 half_precision: bool = True,
                 stream: bool = False,
                 interpolate: bool = False,
                 threshold: float = 0.7,
                 max_tracks: Optional[int] = 2,
                 imgsz: int = 320,
                 iou: float = 0.5):

        if isinstance(video_path, list):
            check_valid_lst(data=video_path, source=f'{self.__class__.__name__} video_path', valid_dtypes=(str, np.str_,), min_len=1)
        elif isinstance(video_path, str):
            check_file_exist_and_readable(file_path=video_path)
            video_path = [video_path]
        for i in video_path: _ = get_video_meta_data(video_path=i)
        check_file_exist_and_readable(file_path=weights_path)
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose')
        check_valid_boolean(value=interpolate, source=f'{self.__class__.__name__} interpolate')
        check_int(name=f'{self.__class__.__name__} batch_size', value=batch_size, min_value=1)
        check_int(name=f'{self.__class__.__name__} imgsz', value=imgsz, min_value=1)
        check_float(name=f'{self.__class__.__name__} threshold', value=threshold, min_value=10e-6, max_value=1.0)
        check_float(name=f'{self.__class__.__name__} iou', value=iou, min_value=10e-6, max_value=1.0)
        check_valid_tuple(x=keypoint_names, source=f'{self.__class__.__name__} keypoint_names', min_integer=1, valid_dtypes=(str,))
        check_file_exist_and_readable(file_path=config_path)
        self.keypoint_col_names = [f'{i}_{s}'.upper() for i in keypoint_names for s in ['x', 'y', 'p']]
        self.keypoint_cord_col_names = [f'{i}_{s}'.upper() for i in keypoint_names for s in ['x', 'y']]
        OUT_COLS.extend(self.keypoint_col_names)
        COORD_COLS.extend(self.keypoint_cord_col_names)
        torch.set_num_threads(torch_threads)
        self.model = load_yolo_model(weights_path=weights_path, device=device, format=format)
        self.half_precision, self.stream, self.video_path, self.config_path = half_precision, stream, video_path, config_path
        self.batch_size, self.threshold, self.max_tracks, self.iou = batch_size, threshold, max_tracks, iou
        self.verbose, self.save_dir, self.imgsz, self.interpolate, self.device = verbose, save_dir, imgsz, interpolate, device
        if self.model.model.task != 'pose':
            raise InvalidFileTypeError(msg=f'The model {weights_path} is not a pose model. It is a {self.model.model.task} model', source=self.__class__.__name__)
        if self.model.model.kpt_shape[0] != len(keypoint_names):
            raise CountError(msg=f'The YOLO model expects {self.model.model.model.head.kpt_shape[0]} keypoints but you passed {len(keypoint_names)}: {keypoint_names}', source=self.__class__.__name__)
        self.class_ids = self.model.names


    def run(self):
        self.results = {}
        timer = SimbaTimer(start=True)
        for video_cnt, video_path in enumerate(self.video_path):
            _, video_name, _ = get_fn_ext(filepath=video_path)
            _ = get_video_meta_data(video_path=video_path)
            video_out = []
            video_predictions = self.model.track(source=video_path, stream=self.stream, tracker=self.config_path, conf=self.threshold, half=self.half_precision, imgsz=self.imgsz, persist=False, iou=self.iou, device=self.device, max_det=self.max_tracks)
            print(video_predictions)
            for frm_cnt, video_prediction in enumerate(video_predictions):
                boxes = video_prediction.obb.data if video_prediction.obb is not None else video_prediction.boxes.data
                boxes = boxes.cpu().numpy().astype(np.float32)
                keypoints = video_prediction.keypoints.data.cpu().numpy().astype(np.float32)
                detected_classes = np.unique(boxes[:, -1]).astype(int) if boxes.size > 0 else []
                for class_id, class_name in self.class_ids.items():
                    if class_id not in detected_classes:
                        video_out.append(_get_undetected_obs(frm_id=frm_cnt, class_id=class_id, class_name=class_name, value_cnt=(10 + (len(self.keypoint_col_names)))))
                        continue
                    if boxes.shape[1] != 7: boxes = np.insert(boxes, 4, -1, axis=1)
                    print(boxes)
                    cls_boxes, cls_keypoints = filter_yolo_keypoint_data(bbox_data=boxes, keypoint_data=keypoints, class_id=class_id, confidence=None, class_idx=-1, confidence_idx=None)
                    for i in range(cls_boxes.shape[0]):
                        frm_results = np.array([frm_cnt, boxes[i][-1], self.class_ids[boxes[i][-1]], boxes[i][-2], boxes[i][-3]])
                        box = np.array([boxes[i][0], boxes[i][1], boxes[i][2], boxes[i][1], boxes[i][2], boxes[i][3], boxes[i][0], boxes[i][3]]).astype(np.int32)
                        frm_results = np.append(frm_results, box)
                        frm_results = np.append(frm_results, keypoints[i].flatten())
                        video_out.append(frm_results)
            self.results[video_name] = pd.DataFrame(video_out, columns=OUT_COLS)
            self.results[video_name][COORD_COLS] = self.results[video_name][COORD_COLS].astype(float).astype(int)
            if self.interpolate:
                for cord_col in COORD_COLS:
                    self.results[video_name][cord_col] = self.results[video_name][cord_col].astype(np.int32).replace(to_replace=-1, value=np.nan)
                    self.results[video_name][cord_col] = self.results[video_name][cord_col].interpolate(method=NEAREST, axis=0).ffill().bfill()
        timer.stop_timer()
        if not self.save_dir:
            if self.verbose:
                print(f'YOLO results created', timer.elapsed_time_str)
            return self.results
        else:
            for k, v in self.results.items():
                save_path = os.path.join(self.save_dir, f'{k}.csv')
                v.to_csv(save_path)
                if self.verbose:
                    print(f'YOLO results saved in {self.save_dir} directory', timer.elapsed_time_str)
        stdout_success(msg=f'{len(self.video_path)} poe-estimation data files saved in {self.save_dir}', elapsed_time=timer.elapsed_time_str)
        return None


# VIDEO_PATH = "/mnt/d/netholabs/yolo_videos/input/mp4_20250606083508/2025-05-28_19-50-23.mp4"
# #VIDEO_PATH = "/mnt/d/netholabs/yolo_videos/2025-05-28_19-46-56.mp4"
# VIDEO_PATH = "/mnt/d/netholabs/yolo_videos/2025-05-28_19-46-56.mp4"
# BOTSORT_PATH = "/mnt/c/projects/simba/simba/simba/assets/bytetrack.yml"
#BOTSORT_PATH = "/mnt/c/projects/simba/simba/simba/assets/botsort.yml"


# VIDEO_PATH = r"/mnt/d/ares/data/termite_2/videos/termite.mp4"
# WEIGHTS_PASS = r"/mnt/d/ares/data/termite_2/yolo/mdl/train13/weights/best.pt"
# SAVE_DIR = "/mnt/d/ares/data/termite_2/yolo/results"

# VIDEO_PATH = r"/mnt/d/ares/data/ant/sleap_video/ant.mp4"
# WEIGHTS_PASS = r"/mnt/d/ares/data/ant/yolo/mdl/train6/weights/best.pt"
# SAVE_DIR = "/mnt/d/ares/data/ant/yolo/results"




# VIDEO_PATH = r"D:\cvat_annotations\videos\mp4_20250624155703\s16-Chasing.mp4"
# WEIGHTS_PASS = r"D:\cvat_annotations\frames\yolo_072125\mdl\train\weights\best.pt"
# SAVE_DIR = r"D:\cvat_annotations\frames\yolo_072125\results_track"
# BOTSORT_PATH = r"C:\projects\simba\simba\simba\assets\bytetrack.yml"
#
# KEYPOINT_NAMES = ('Nose', 'Left_ear', 'Right_ear', 'Left_side', 'Center', 'Right_side', 'Tail_base', 'Tail_center', 'Tail_tip')
#
# i = YOLOPoseTrackInference(weights_path=WEIGHTS_PASS,
#                            video_path=VIDEO_PATH,
#                            save_dir=SAVE_DIR,
#                            verbose=True,
#                            device=0,
#                            format=None,
#                            keypoint_names=KEYPOINT_NAMES,
#                            batch_size=32,
#                            threshold=0.01,
#                            config_path=BOTSORT_PATH,
#                            interpolate=False,
#                            imgsz=640,
#                            max_tracks=5,
#                            stream=False,
#                            iou=0.2)
# i.run()



__author__ = "Simon Nilsson"

import os
from copy import deepcopy
from typing import Any, Dict, List, Optional, Union

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import numpy as np
import pandas as pd

from simba.data_processors.interpolate import Interpolate
from simba.data_processors.smoothing import Smoothing
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pose_importer_mixin import PoseImporterMixin
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists,
                                check_if_keys_exist_in_dict, check_instance,
                                check_int, check_str, check_valid_lst)
from simba.utils.enums import Formats, Methods, Options
from simba.utils.errors import BodypartColumnNotFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_all_videos_in_project,
                                    get_video_meta_data, write_df)


class MADLCImporterH5(ConfigReader, PoseImporterMixin):
    """
    Importing multi-animal deeplabcut (maDLC) pose-estimation data (in H5 format)
    into a SimBA project in parquet or CSV format.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str data_folder: Path to folder containing maDLC data in ``.h5`` format.
    :parameter str file_type: Method used to perform pose-estimation in maDLC. OPTIONS: `skeleton`, `box`, `ellipse`.
    :param List[str] id_lst: Names of animals.
    :parameter Optional[Dict[str, str]] interpolation_setting: Dict defining the type and method to use to perform interpolation {'type': 'animals', 'method': 'linear'}.
    :parameter Optional[Dict[str, Union[str, int]]] smoothing_settings: Dictionary defining the pose estimation smoothing method {'time_window': 500, 'method': 'gaussian'}.

    .. note::
       `Multi-animal import tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Multi_animal_pose.md>`__.

    :examples:
    >>> _ = MADLCImporterH5(config_path=r'MyConfigPath', data_folder=r'maDLCDataFolder', file_type='ellipse', id_lst=['Animal_1', 'Animal_2'], interpolation_settings={'type': 'animals', 'method': 'linear'}, smoothing_settings={'time_window': 500, 'method': 'gaussian'}).run()

    References
    ----------
    .. [1] Lauer et al., Multi-animal pose estimation, identification and tracking with DeepLabCut, `Nature Methods`,
           2022.
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_folder: Union[str, os.PathLike],
                 file_type: Literal['skeleton', 'box', 'ellipse'],
                 id_lst: List[str],
                 interpolation_settings: Optional[Dict[str, str]] = None,
                 smoothing_settings: Optional[Dict[str, Any]] = None):

        check_file_exist_and_readable(file_path=config_path)
        check_if_dir_exists(in_dir=data_folder)
        check_str(name=f'{self.__class__.__name__} file_type', value=file_type, options=Options.MULTI_DLC_TYPE_IMPORT_OPTION.value)
        check_valid_lst(data=id_lst, source=f'{self.__class__.__name__} id_lst', valid_dtypes=(str,))
        if interpolation_settings is not None:
            check_if_keys_exist_in_dict(data=interpolation_settings, key=['method', 'type'], name=f'{self.__class__.__name__} interpolation_settings')
            check_str(name=f'{self.__class__.__name__} interpolation_settings type', value=interpolation_settings['type'], options=('body-parts', 'animals'))
            check_str(name=f'{self.__class__.__name__} interpolation_settings method', value=interpolation_settings['method'], options=('linear', 'quadratic', 'nearest'))
        if smoothing_settings is not None:
            check_if_keys_exist_in_dict(data=smoothing_settings, key=['method', 'time_window'], name=f'{self.__class__.__name__} smoothing_settings')
            check_str(name=f'{self.__class__.__name__} smoothing_settings method', value=smoothing_settings['method'], options=('savitzky-golay', 'gaussian'))
            check_int(name=f'{self.__class__.__name__} smoothing_settings time_window', value=smoothing_settings['time_window'], min_value=1)

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        PoseImporterMixin.__init__(self)
        self.interpolation_settings, self.smoothing_settings = interpolation_settings, smoothing_settings
        self.data_folder, self.id_lst = data_folder, id_lst
        self.import_log_path = os.path.join(self.logs_path, f"data_import_log_{self.datetime}.csv")
        self.video_paths = find_all_videos_in_project(videos_dir=self.video_dir, raise_error=True if len(id_lst) > 1 else False)
        self.input_data_paths = self.find_data_files(dir=self.data_folder, extensions=Formats.DLC_FILETYPES.value[file_type])
        self.data_and_videos_lk = self.link_video_paths_to_data_paths(data_paths=self.input_data_paths, video_paths=self.video_paths, str_splits=Formats.DLC_NETWORK_FILE_NAMES.value, raise_error=True if len(id_lst) > 1 else False)
        if self.pose_setting is Methods.USER_DEFINED.value:
            self.__update_config_animal_cnt()
        if self.animal_cnt > 1:
            self.check_multi_animal_status()
            self.animal_bp_dict = self.create_body_part_dictionary(self.multi_animal_status, self.id_lst, self.animal_cnt, self.x_cols, self.y_cols, self.p_cols, self.clr_lst)
            if self.pose_setting is Methods.USER_DEFINED.value:
                self.update_bp_headers_file(update_bp_headers=True)
        print(f"Importing {len(list(self.data_and_videos_lk.keys()))} multi-animal DLC file(s)...")

    def run(self):
        import_log = pd.DataFrame(columns=["VIDEO", "IMPORT_TIME", "IMPORT_SOURCE", "INTERPOLATION_SETTING", "SMOOTHING_SETTING"])
        for cnt, (video_name, video_data) in enumerate(self.data_and_videos_lk.items()):
            video_timer = SimbaTimer(start=True)
            self.add_spacer, self.frame_no, self.video_data, self.video_name = (2, 1, video_data, video_name)
            print(f"Processing {video_name} ({cnt+1}/{len(self.input_data_paths)})...")
            self.data_df = pd.read_hdf(video_data["DATA"]).replace([np.inf, -np.inf], np.nan).fillna(0)
            if len(self.data_df.columns) != len(self.bp_headers):
                raise BodypartColumnNotFoundError(
                    msg=f'The number of body-parts in data file {video_data["DATA"]} do not match the number of body-parts in your SimBA project. '
                    f"The number of of body-parts expected by your SimBA project is {int(len(self.bp_headers) / 3)}. "
                    f'The number of of body-parts contained in data file {video_data["DATA"]} is {int(len(self.data_df.columns) / 3)}. '
                    f"Make sure you have specified the correct number of animals and body-parts in your project. NOTE: The project body-parts is stored at {self.body_parts_path}."
                )
            self.data_df.columns = self.bp_headers
            if self.animal_cnt > 1:
                self.initialize_multi_animal_ui(animal_bp_dict=self.animal_bp_dict, video_info=get_video_meta_data(video_data["VIDEO"]), data_df=self.data_df, video_path=video_data["VIDEO"])
                self.multianimal_identification()
            else:
                self.out_df = self.insert_multi_idx_columns(df=self.data_df.fillna(0))
            self.save_path = os.path.join(os.path.join(self.input_csv_dir, f"{self.video_name}.{self.file_type}"))
            write_df(df=self.out_df, file_type=self.file_type, save_path=self.save_path, multi_idx_header=True)
            if self.interpolation_settings is not None:
                interpolator = Interpolate(config_path=self.config_path, data_path=self.save_path, type=self.interpolation_settings['type'], method=self.interpolation_settings['method'], multi_index_df_headers=True, copy_originals=False)
                interpolator.run()
            if self.smoothing_settings is not None:
                smoother = Smoothing(config_path=self.config_path, data_path=self.save_path, time_window=self.smoothing_settings['time_window'], method=self.smoothing_settings['method'], multi_index_df_headers=True, copy_originals=False)
                smoother.run()
            video_timer.stop_timer()
            stdout_success(msg=f"Video {video_name} data imported...", elapsed_time=video_timer.elapsed_time_str)
        self.timer.stop_timer()
        stdout_success(msg=f"All maDLC H5 data files imported to {self.input_csv_dir} directory", elapsed_time=self.timer.elapsed_time_str)


# test = MADLCImporterH5(config_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                    data_folder=r'/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/h5',
#                    file_type='ellipse',
#                    id_lst=['Simon', 'JJ'],
#                    interpolation_settings= {'type': 'animals', 'method': 'linear'},
#                    smoothing_settings = {'time_window': 500, 'method': 'gaussian'})
# test.run()



import functools
import multiprocessing
import os
import platform
from pathlib import Path
from typing import Dict, Optional, Union

import cv2
import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.geometry_mixin import GeometryMixin
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (check_instance, check_int,
                                check_nvidea_gpu_available, check_str,
                                check_that_column_exist, check_valid_boolean)
from simba.utils.data import create_color_palette
from simba.utils.enums import OS, Formats, Options
from simba.utils.errors import CountError, InvalidFilepathError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    find_core_cnt,
                                    find_files_of_filetypes_in_directory,
                                    get_fn_ext, get_video_meta_data, read_df)
from simba.utils.warnings import FrameRangeWarning


def pose_plotter_mp(data: pd.DataFrame,
                    video_meta_data: dict,
                    video_path: str,
                    bp_dict: dict,
                    colors_dict: dict,
                    circle_size: int,
                    bbox: bool,
                    video_save_dir: Union[str, os.PathLike],):

    fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
    group_cnt = int(data.iloc[0]["group"])
    data = data.drop(["group"], axis=1)
    start_frm, current_frm, end_frm = data.index[0], data.index[0], data.index[-1]
    save_path = os.path.join(video_save_dir, f"{group_cnt}.mp4")
    writer = cv2.VideoWriter(save_path, fourcc, video_meta_data["fps"], (video_meta_data["width"], video_meta_data["height"]))
    cap = cv2.VideoCapture(video_path)
    cap.set(1, start_frm)
    while current_frm < end_frm:
        ret, img = cap.read()
        if ret:
            for animal_cnt, (animal_name, animal_data) in enumerate(bp_dict.items()):
                animal_bbox = []
                for cnt, (x_name, y_name) in enumerate(zip(animal_data["X_bps"], animal_data["Y_bps"])):
                    check_that_column_exist(df=data, column_name=[x_name, y_name], file_name=video_path)
                    bp_tuple = (int(data.at[current_frm, x_name]), int(data.at[current_frm, y_name]))
                    clr = colors_dict[animal_cnt][cnt]
                    img = cv2.circle(img, bp_tuple, circle_size, clr, -1)
                    animal_bbox.append(list(bp_tuple))
                if bbox and len(animal_bbox) > 4:
                    animal_bbox = GeometryMixin().keypoints_to_axis_aligned_bounding_box(keypoints=np.array(animal_bbox).reshape(-1, len(animal_bbox), 2).astype(np.int32))
                    img = cv2.polylines(img, [animal_bbox], True, colors_dict[animal_cnt][0], thickness=max(1, int(circle_size/1.5)), lineType=-1)
            writer.write(img)
            current_frm += 1
            print(f"Multi-processing video frame {current_frm} on core {group_cnt}...")
        else:
            print(f'Frame {current_frm} not found in video {video_path}, terminating video creation...')
            break
    cap.release()
    writer.release()


class PosePlotterMultiProcess():
    """
    Create pose-estimation visualizations from data within a SimBA project folder.

    :param str in_directory: Path to SimBA project directory containing pose-estimation data in parquet or CSV format.
    :param str out_directory: Directory to where to save the pose-estimation videos.
    :param int Size of the circles denoting the location of the pose-estimated body-parts.
    :param Optional[dict] clr_attr: Python dict where animals are keys and color attributes values. E.g., {'Animal_1':  (255, 107, 198)}. If None, random palettes will be used.

    .. image:: _static/img/pose_plotter.png
       :width: 600
       :align: center

    :example:
    >>> test = PosePlotterMultiProcess(in_dir='project_folder/csv/input_csv', out_dir='/project_folder/test_viz', circle_size=10, core_cnt=1, color_settings={'Animal_1':  'Green', 'Animal_2':  'Red'})
    >>> test.run()
    """

    def __init__(self,
                 data_path: Union[str, os.PathLike],
                 out_dir: Optional[Union[str, os.PathLike]] = None,
                 palettes: Optional[Dict[str, str]] = None,
                 circle_size: Optional[int] = None,
                 core_cnt: Optional[int] = -1,
                 gpu: Optional[bool] = False,
                 bbox: Optional[bool] = False,
                 sample_time: Optional[int] = None) -> None:

        if os.path.isdir(data_path):
            config_path = os.path.join(Path(data_path).parents[1], 'project_config.ini')
        elif os.path.isfile(data_path):
            config_path = os.path.join(Path(data_path).parents[2], 'project_config.ini')
        else:
            raise InvalidFilepathError(msg=f'{data_path} not not a valid file or directory path.', source=self.__class__.__name__)
        if not os.path.isfile(config_path):
            raise InvalidFilepathError(msg=f'When visualizing pose-estimation, select an input sub-directory of the project_folder/csv folder OR file in the project_folder/csv/ANY_FOLDER directory. {data_path} does not meet these requirements and therefore SimBA cant locate the project_config.ini (expected at {config_path}', source=self.__class__.__name__)
        self.config = ConfigReader(config_path=config_path, read_video_info=False, create_logger=False)
        if os.path.isdir(data_path):
            files_found = find_files_of_filetypes_in_directory(directory=data_path, extensions=[f'.{self.config.file_type}'], raise_error=True)
        else:
            files_found = [data_path]
        self.animal_bp_dict = self.config.body_parts_lst
        if circle_size is not None:
            check_int(name='circle_size', value=circle_size, min_value=1)
        check_int(name='core_cnt', value=core_cnt, min_value=-1)
        if core_cnt == -1: core_cnt = find_core_cnt()[0]
        self.color_dict = {}
        if palettes is not None:
            check_instance(source=self.__class__.__name__, instance=palettes, accepted_types=(dict,))
            if len(list(palettes.keys())) != self.config.animal_cnt:
                raise CountError(msg=f'The number of color palettes ({(len(list(palettes.keys())))}) spedificed is not the same as the number of animals ({(self.config.animal_cnt)}) in the SimBA project at {self.config.project_path}')
            for cnt, (k, v) in enumerate(palettes.items()):
                check_str(name='palette', value=v, options=Options.PALETTE_OPTIONS_CATEGORICAL.value)
                self.color_dict[cnt] = create_color_palette(pallete_name=v, increments=len(self.config.body_parts_lst))
        else:
            for cnt, (k, v) in enumerate(self.config.animal_bp_dict.items()):
                self.color_dict[cnt] = self.config.animal_bp_dict[k]["colors"]
        check_valid_boolean(value=bbox, source=f'{self.__class__.__name__} bbox')
        if sample_time is not None:
            check_int(name='sample_time', value=sample_time, min_value=1)
        if out_dir is None:
            out_dir = os.path.join(os.path.dirname(files_found[0]), f'pose_videos_{self.config.datetime}')
        self.circle_size, self.core_cnt, self.out_dir, self.sample_time, self.bbox = (circle_size, core_cnt, out_dir, sample_time, bbox)
        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)
        self.data = {}
        check_valid_boolean(value=[gpu])

        if gpu and check_nvidea_gpu_available():
            self.gpu = True
        else:
            self.gpu = False
        for file in files_found:
            self.data[file] = self.config.find_video_of_file(video_dir=self.config.video_dir, filename=get_fn_ext(file)[1])
        if platform.system() == OS.MAC.value:
            multiprocessing.set_start_method("spawn", force=True)

    def run(self):
        for file_cnt, (pose_path, video_path) in enumerate(self.data.items()):
            video_timer = SimbaTimer(start=True)
            video_name = get_fn_ext(pose_path)[1]
            self.temp_folder = os.path.join(self.out_dir, video_name, "temp")
            if os.path.exists(self.temp_folder):
                self.config.remove_a_folder(self.temp_folder)
            os.makedirs(self.temp_folder, exist_ok=True)
            save_video_path = os.path.join(self.out_dir, f"{video_name}.mp4")
            pose_df = read_df(file_path=pose_path, file_type=self.config.file_type, check_multiindex=True)
            video_meta_data = get_video_meta_data(video_path=video_path)
            if self.circle_size is None:
                video_circle_size = PlottingMixin().get_optimal_circle_size(frame_size=(int(video_meta_data['width']), int(video_meta_data['height'])), circle_frame_ratio=70)
            else:
                video_circle_size = self.circle_size
            if (self.sample_time is None) and (video_meta_data["frame_count"] != len(pose_df)):
                FrameRangeWarning(msg=f'The video {video_name} has pose-estimation data for {len(pose_df)} frames, but the video has {video_meta_data["frame_count"]} frames. Ensure the data and video has an equal number of frames.', source=self.__class__.__name__)
            elif isinstance(self.sample_time, int):
                sample_frm_cnt = int(video_meta_data["fps"] * self.sample_time)
                if sample_frm_cnt > len(pose_df): sample_frm_cnt = len(pose_df)
                pose_df = pose_df.iloc[0:sample_frm_cnt]
            if 'input_csv' in os.path.dirname(pose_path):
                pose_df = self.config.insert_column_headers_for_outlier_correction(data_df=pose_df, new_headers=self.config.bp_headers, filepath=pose_path)
            pose_df = (pose_df.apply(pd.to_numeric, errors="coerce").fillna(0).reset_index(drop=True))
            pose_lst, obs_per_split = PlottingMixin().split_and_group_df(df=pose_df, splits=self.core_cnt)
            print(f"Creating pose videos, multiprocessing (chunksize: {self.config.multiprocess_chunksize}, cores: {self.core_cnt})...")
            with multiprocessing.Pool(self.core_cnt, maxtasksperchild=self.config.maxtasksperchild) as pool:
                constants = functools.partial(pose_plotter_mp,
                                              video_meta_data=video_meta_data,
                                              video_path=video_path,
                                              bp_dict=self.config.animal_bp_dict,
                                              colors_dict=self.color_dict,
                                              circle_size=video_circle_size,
                                              bbox=self.bbox,
                                              video_save_dir=self.temp_folder)
                for cnt, result in enumerate(pool.imap(constants, pose_lst, chunksize=self.config.multiprocess_chunksize)):
                    print(f"Image {min(len(pose_df), obs_per_split*(cnt+1))}/{len(pose_df)}, Video {file_cnt+1}/{len(list(self.data.keys()))}...")
            pool.terminate()
            pool.join()
            print(f"Joining {video_name} multi-processed video...")
            concatenate_videos_in_folder(in_folder=self.temp_folder, save_path=save_video_path, remove_splits=True, gpu=self.gpu)
            video_timer.stop_timer()
            stdout_success(msg=f"Pose video {video_name} complete and saved at {save_video_path}", elapsed_time=video_timer.elapsed_time_str, source=self.__class__.__name__)
        self.config.timer.stop_timer()
        stdout_success(f"Pose visualizations for {len(list(self.data.keys()))} video(s) created in {self.out_dir} directory", elapsed_time=self.config.timer.elapsed_time_str, source=self.__class__.__name__)


if __name__ == "__main__":
    test = PosePlotterMultiProcess(data_path=r"C:\troubleshooting\mitra\project_folder\csv\input_csv\501_MA142_Gi_DCZ_0603.csv",
                                   out_dir=None,
                                   circle_size=8,
                                   core_cnt=18,
                                   palettes=None,
                                   bbox=True,)
    test.run()




# test = PosePlotterMultiProcess(data_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/outlier_corrected_movement_location/Together_1.csv',
#                    out_dir=None,
#                    circle_size=None,
#                    core_cnt=1,
#                    palettes=None)
# test.run()


# test = PosePlotter(in_dir='/Users/simon/Desktop/envs/troubleshooting/dam_nest-c-only_ryan/project_folder/csv/outlier_corrected_movement_location',
#                    out_dir='/Users/simon/Desktop/video_tests_',
#                    sample_time=2,
#                    circle_size=10,
#                    core_cnt=1,
#                    color_settings=None) #
# test.run()


# test = PosePlotter(in_dir='/Users/simon/Desktop/envs/troubleshooting/piotr/project_folder/csv/outlier_corrected_movement_location',
#                    out_dir='/Users/simon/Desktop/envs/troubleshooting/piotr/project_folder/frames/output/test',
#                    circle_size=10,
#                    core_cnt=6,
#                    color_settings={'Animal_1':  'Green', 'Animal_2':  'Red'})
# test.run()



__author__ = "Simon Nilsson"

import itertools
import os
from typing import Optional, Union

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_that_dir_has_list_of_filenames)
from simba.utils.enums import TagNames
from simba.utils.errors import AnimalNumberError, CountError, InvalidInputError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (get_fn_ext, read_data_paths, read_df,
                                    write_df)


class DirectingOtherAnimalsAnalyzer(ConfigReader, FeatureExtractionMixin):
    """
    Calculate when animals are directing towards body-parts of other animals. Results are stored in
    the ``project_folder/logs/directionality_dataframes`` directory of the SimBA project.

    .. note:
       `Example expected bool table <https://github.com/sgoldenlab/simba/blob/master/misc/boolean_directionaly_example.csv>`__.
       `Example expected summary table <https://github.com/sgoldenlab/simba/blob/master/misc/detailed_summary_directionality_example.csv>`__.
       `Example expected aggregate statistics table <https://github.com/sgoldenlab/simba/blob/master/misc/direction_data_aggregates_example.csv>`__.

    .. important::
       Requires the pose-estimation data for the ``left ear``, ``right ear`` and ``nose`` of each individual animals.
       `Github Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#part-3-generating-features-from-roi-data>`__.
       `Expected output <https://github.com/sgoldenlab/simba/blob/master/misc/Direction_data_example.csv>`__.

    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter bool bool_tables: If True, creates boolean output tables.
    :parameter bool summary_tables: If True, creates summary tables including approximate location of eye of observer and the location of observed body-parts and frames when observation was detected.
    :parameter bool aggregate_statistics_tables: If True, summary statistics tables of how much time each animal spent observation the other animals.

    :examples:
    >>> directing_analyzer = DirectingOtherAnimalsAnalyzer(config_path='MyProjectConfig')
    >>> directing_analyzer.run()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        data_paths: Optional[Union[str, os.PathLike, None]] = None,
        bool_tables: Optional[bool] = True,
        summary_tables: Optional[bool] = False,
        append_bool_tables_to_features: Optional[bool] = False,
        aggregate_statistics_tables: Optional[bool] = False,
    ):

        check_file_exist_and_readable(file_path=config_path)
        ConfigReader.__init__(self, config_path=config_path)
        FeatureExtractionMixin.__init__(self, config_path=config_path)
        data_paths = read_data_paths(
            path=data_paths,
            default=self.outlier_corrected_paths,
            default_name=self.outlier_corrected_dir,
            file_type=self.file_type,
        )
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        if self.animal_cnt < 2:
            raise AnimalNumberError(
                "Cannot analyze directionality between animals in a project with less than two animals.",
                source=self.__class__.__name__,
            )
        self.animal_permutations = list(itertools.permutations(self.animal_bp_dict, 2))
        (
            self.bool_tables,
            self.summary_tables,
            self.aggregate_statistics_tables,
            self.append_bool_tables_to_features,
        ) = (
            bool_tables,
            summary_tables,
            aggregate_statistics_tables,
            append_bool_tables_to_features,
        )
        self.data_paths = data_paths
        if self.append_bool_tables_to_features:
            check_that_dir_has_list_of_filenames(
                dir=self.features_dir,
                file_name_lst=self.outlier_corrected_paths,
                file_type=self.file_type,
            )
        print(f"Processing {len(self.data_paths)} video(s)...")
        if not self.check_directionality_viable()[0]:
            raise InvalidInputError(
                msg="You are not tracking the necessary body-parts to calculate direction.",
                source=self.__class__.__name__,
            )

    #
    def run(self):
        if self.aggregate_statistics_tables:
            check_all_file_names_are_represented_in_video_log(
                video_info_df=self.video_info_df, data_paths=self.data_paths
            )
        self.results = {}
        for file_cnt, file_path in enumerate(self.data_paths):
            video_timer = SimbaTimer(start=True)
            _, video_name, _ = get_fn_ext(file_path)
            self.results[video_name] = {}
            print(f"Analyzing directionality between animals in video {video_name}...")
            data_df = read_df(file_path=file_path, file_type=self.file_type)
            direct_bp_dict = self.check_directionality_cords()
            for animal_permutation in self.animal_permutations:
                self.results[video_name][
                    f"{animal_permutation[0]} directing towards {animal_permutation[1]}"
                ] = {}
                first_animal_bps, second_animal_bps = (
                    direct_bp_dict[animal_permutation[0]],
                    self.animal_bp_dict[animal_permutation[1]],
                )
                first_ear_left_arr = data_df[
                    [
                        first_animal_bps["Ear_left"]["X_bps"],
                        first_animal_bps["Ear_left"]["Y_bps"],
                    ]
                ].values
                first_ear_right_arr = data_df[
                    [
                        first_animal_bps["Ear_right"]["X_bps"],
                        first_animal_bps["Ear_right"]["Y_bps"],
                    ]
                ].values
                first_nose_arr = data_df[
                    [
                        first_animal_bps["Nose"]["X_bps"],
                        first_animal_bps["Nose"]["Y_bps"],
                    ]
                ].values
                second_animal_x_bps, second_animal_y_bps = (
                    second_animal_bps["X_bps"],
                    second_animal_bps["Y_bps"],
                )
                for x_bp, y_bp in zip(second_animal_x_bps, second_animal_y_bps):
                    target_cord_arr = data_df[[x_bp, y_bp]].values
                    direction_data = self.jitted_line_crosses_to_nonstatic_targets(
                        left_ear_array=first_ear_left_arr,
                        right_ear_array=first_ear_right_arr,
                        nose_array=first_nose_arr,
                        target_array=target_cord_arr,
                    )
                    x_min = np.minimum(direction_data[:, 1], first_nose_arr[:, 0])
                    y_min = np.minimum(direction_data[:, 2], first_nose_arr[:, 1])
                    delta_x = abs((direction_data[:, 1] - first_nose_arr[:, 0]) / 2)
                    delta_y = abs((direction_data[:, 2] - first_nose_arr[:, 1]) / 2)
                    x_middle, y_middle = np.add(x_min, delta_x), np.add(y_min, delta_y)
                    direction_data = np.concatenate(
                        (y_middle.reshape(-1, 1), direction_data), axis=1
                    )
                    direction_data = np.concatenate(
                        (x_middle.reshape(-1, 1), direction_data), axis=1
                    )
                    direction_data = np.delete(direction_data, [2, 3, 4], 1)
                    direction_data = np.hstack((direction_data, target_cord_arr))
                    bp_data = pd.DataFrame(
                        direction_data,
                        columns=["Eye_x", "Eye_y", "Directing_BOOL", x_bp, y_bp],
                    )
                    bp_data = bp_data[["Eye_x", "Eye_y", x_bp, y_bp, "Directing_BOOL"]]
                    bp_data.insert(loc=0, column="Animal_2_body_part", value=x_bp[:-2])
                    bp_data.insert(
                        loc=0, column="Animal_2", value=animal_permutation[1]
                    )
                    bp_data.insert(
                        loc=0, column="Animal_1", value=animal_permutation[0]
                    )
                    self.results[video_name][
                        f"{animal_permutation[0]} directing towards {animal_permutation[1]}"
                    ][x_bp[:-2]] = bp_data
            video_timer.stop_timer()
            print(
                f"Direction analysis complete for video {video_name} ({file_cnt + 1}/{len(self.outlier_corrected_paths)}, elapsed time: {video_timer.elapsed_time_str}s)..."
            )

        if self.bool_tables:
            save_dir = os.path.join(
                self.logs_path, f"Animal_directing_animal_booleans_{self.datetime}"
            )
            if not os.path.isdir(save_dir):
                os.makedirs(save_dir)
            for video_cnt, (video_name, video_data) in enumerate(self.results.items()):
                print(
                    f"Saving boolean directing tables for video {video_name} (Video {video_cnt + 1}/{len(self.results.keys())})..."
                )
                video_df = pd.DataFrame()
                for animal_permutation, animal_permutation_data in video_data.items():
                    for (
                        body_part_name,
                        body_part_data,
                    ) in animal_permutation_data.items():
                        video_df[f"{animal_permutation}_{body_part_name}"] = (
                            body_part_data["Directing_BOOL"]
                        )
                if self.append_bool_tables_to_features:
                    print(f"Adding directionality tables to features data for video {video_name}...")
                    df = read_df(
                        file_path=os.path.join(
                            self.features_dir, f"{video_name}.{self.file_type}"
                        ),
                        file_type=self.file_type,
                    )
                    if len(df) != len(video_df):
                        raise CountError(
                            msg=f"Failed to join data files as they contains different number of frames: the file representing video {video_name} in directory {self.outlier_corrected_dir} contains {len(video_df)} frames, and the file representing video {video_name} in directory {self.features_dir} contains {len(df)} frames."
                        )
                    else:
                        df = pd.concat(
                            [
                                df.reset_index(drop=True),
                                video_df.reset_index(drop=True),
                            ],
                            axis=1,
                        )
                    write_df(
                        df=df,
                        file_type=self.file_type,
                        save_path=os.path.join(
                            self.features_dir, f"{video_name}.{self.file_type}"
                        ),
                    )
                video_df.to_csv(os.path.join(save_dir, f"{video_name}.csv"))
            stdout_success(
                msg=f"All boolean tables saved in {save_dir}!",
                source=self.__class__.__name__,
            )

        if self.aggregate_statistics_tables:
            print("Computing summary statistics...")
            save_path = os.path.join(
                self.logs_path, f"Direction_aggregate_summary_data_{self.datetime}.csv"
            )
            out_df = pd.DataFrame(columns=["VIDEO", "ANIMAL PERMUTATION", "VALUE (S)"])
            for video_name, video_data in self.results.items():
                _, _, fps = self.read_video_info(video_name=video_name)
                for animal_permutation, permutation_data in video_data.items():
                    idx_directing = set()
                    for bp_name, bp_data in permutation_data.items():
                        idx_directing.update(
                            list(bp_data.index[bp_data["Directing_BOOL"] == 1])
                        )
                    value = round(len(idx_directing) / fps, 3)
                    out_df.loc[len(out_df)] = [video_name, animal_permutation, value]

            self.out_df = out_df.sort_values(
                by=["VIDEO", "ANIMAL PERMUTATION"]
            ).set_index("VIDEO")
            self.out_df.to_csv(save_path)
            stdout_success(
                msg=f"Summary directional statistics saved at {save_path}",
                source=self.__class__.__name__,
            )

        if self.summary_tables:
            self.transpose_results()
            save_dir = os.path.join(
                self.logs_path,
                f"detailed_directionality_summary_dataframes_{self.datetime}",
            )
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            for video_cnt, (video_name, video_data) in enumerate(
                self.directionality_df_dict.items()
            ):
                save_name = os.path.join(save_dir, f"{video_name}.csv")
                video_data.to_csv(save_name)
            stdout_success(
                f"All detailed directional data saved in the {save_dir} directory!",
                source=self.__class__.__name__,
            )
        self.timer.stop_timer()
        stdout_success(
            msg="All directional data saved in SimBA project",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )

    def transpose_results(self):
        self.directionality_df_dict = {}
        for video_name, video_data in self.results.items():
            out_df_lst = []
            for animal_permutation, permutation_data in video_data.items():
                for bp_name, bp_data in permutation_data.items():
                    directing_df = (
                        bp_data[bp_data["Directing_BOOL"] == 1]
                        .reset_index()
                        .rename(
                            columns={
                                "index": "Frame_#",
                                bp_name + "_x": "Animal_2_bodypart_x",
                                bp_name + "_y": "Animal_2_bodypart_y",
                            }
                        )
                    )
                    directing_df.insert(loc=0, column="Video", value=video_name)
                    out_df_lst.append(directing_df)
            self.directionality_df_dict[video_name] = pd.concat(
                out_df_lst, axis=0
            ).drop("Directing_BOOL", axis=1)


# test = DirectingOtherAnimalsAnalyzer(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                      bool_tables=True,
#                                      summary_tables=True,
#                                      aggregate_statistics_tables=True,
#                                      append_bool_tables_to_features=False,
#                                      data_paths=None)
# test.run()



import argparse
import os
import warnings
from typing import List, Optional, Union

import numpy as np
import pandas as pd

warnings.filterwarnings("ignore")

import multiprocessing
import platform

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_supplement_mixin import \
    FeatureExtractionSupplemental
from simba.mixins.geometry_mixin import GeometryMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log, check_float, check_int,
    check_str, check_valid_lst, check_video_has_rois)
from simba.utils.errors import (AnimalNumberError, InvalidInputError,
                                NoFilesFoundError, ROICoordinatesNotFoundError)
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_data_paths, read_df
from simba.utils.warnings import NoFileFoundWarning

TAIL_END = "tail_end"

class SpontaneousAlternationCalculator(ConfigReader):
    """
    Compute spontaneous alternations based on specified ROIs and animal detection parameters.

    .. note::
       This method computes spontaneous alternation by fitting the smallest viable geometry (a.k.a. shape, polygon) that
       encompasses the animal key-points (but see buffer parameter below). It then checks the percent overlap between the
       animal geometry and the defined arm and center geometries in each frame. If the percent overlap is more or equal to the
       specified threshold, then the animal considered to visiting the relevant arm. The animal is considered exiting an arm
       when the percent overlap with a different ROI is above the threshold.

    .. attention::
       Requires SimBA project with (i) only one tracked animal, (ii) at least three pose-estmated body-parts, and (iii) defined
       ROIs representing the arms and the center of the maze.

    :param Union[str, os.PathLike] config_path: Path to SimBA project config file.
    :param List[str] arm_names: List of ROI names representing the arms.
    :param str center_name: Name of the ROI representing the center of the maze
    :param Optional[int] animal_area: Value between 51 and 100, representing the percent of the animal body that has to be situated in a ROI for it to be considered an entry.
    :param Optional[float] threshold: Value between 0.0 and 1.0. Body-parts with detection probabilities below this value will be (if possible) filtered when constructing the animal geometry.
    :param Optional[int] buffer: Millimeters area for which the animal geometry should be increased in size. Useful if the animal geometry does not fully cover the animal.
    :param Optional[bool] detailed_data: If True, saves an additional CSV for each analyzed video with detailed data pertaining each error type and alternation sequence.
    :param Optional[Union[str, os.PathLike]] data_path: Directory of path to the file to be analyzed. If None, then ``project_folder/outlier_corrected_movement_location`` directory.

    :example:
    >>> x = SpontaneousAlternationCalculator(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/spontenous_alternation/project_folder/project_config.ini', arm_names=['A', 'B', 'C'], center_name='Center', threshold=0.0, animal_area=100, buffer=2, detailed_data=True)
    >>> x.run()
    >>> x.save()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 arm_names: List[str],
                 center_name: str,
                 animal_area: Optional[int] = 80,
                 threshold: Optional[float] = 0.0,
                 buffer: Optional[int] = 2,
                 verbose: Optional[bool] = False,
                 detailed_data: Optional[bool] = False,
                 data_path: Optional[Union[str, os.PathLike]] = None):

        ConfigReader.__init__(self, config_path=config_path)
        if self.animal_cnt != 1:
            raise AnimalNumberError(msg=f"Spontaneous alternation can only be calculated in 1 animal projects. Your project has {self.animal_cnt} animals.", source=self.__class__.__name__)
        if len(self.body_parts_lst) < 3:
            raise InvalidInputError(msg=f"Spontaneous alternation can only be calculated in projects with 3 or more tracked body-parts. Found {len(self.body_parts_lst)}.", source=self.__class__.__name__)
        if not os.path.isfile(self.roi_coordinates_path):
            raise ROICoordinatesNotFoundError(expected_file_path=self.roi_coordinates_path)
        check_valid_lst(data=arm_names, source=SpontaneousAlternationCalculator.__name__, valid_dtypes=(str,), min_len=2)
        check_str(name="CENTER NAME", value=center_name)
        check_int(name="ANIMAL AREA", value=animal_area, min_value=51, max_value=100)
        check_float(name="THRESHOLD", value=threshold, min_value=0.0, max_value=1.0)
        check_int(name="BUFFER", value=buffer, min_value=1)
        data_paths = read_data_paths(path=data_path, default=self.outlier_corrected_paths, default_name=self.outlier_corrected_dir, file_type=self.file_type)
        file_paths = {}
        for file_path in data_paths: file_paths[get_fn_ext(file_path)[1]] = file_path
        self.read_roi_data()
        files_w_missing_rois = list(set(file_paths.keys()) - set(self.video_names_w_rois))
        self.files_w_rois = [x for x in list(file_paths.keys()) if x in self.video_names_w_rois]
        if len(self.files_w_rois) == 0:
            raise NoFilesFoundError(msg=f"No ROI definitions found for any of the data fat {arm_names}", source=__class__.__name__,)
        if len(files_w_missing_rois) > 0:
            NoFileFoundWarning(msg=f"{len(files_w_missing_rois)} file(s) at {data_path} are missing ROI definitions and will be skipped when performing spontaneous alternation calculations: {files_w_missing_rois}", source=__class__.__name__)
        check_video_has_rois(roi_dict=self.roi_dict, video_names=self.files_w_rois, roi_names=arm_names + [center_name],)
        self.file_paths = list({file_paths[k] for k in self.files_w_rois if k in file_paths})
        self.threshold, self.buffer, self.animal_area, self.detailed_data = threshold, buffer, animal_area, detailed_data
        self.verbose, self.center_name, self.arm_names = verbose, center_name, arm_names

    def run(self):
        roi_geos, self.roi_clrs = GeometryMixin.simba_roi_to_geometries(rectangles_df=self.rectangles_df, circles_df=self.circles_df, polygons_df=self.polygon_df, color=True)
        self.roi_geos = {k: v for k, v in roi_geos.items() if k in self.files_w_rois}
        self.results = {}
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.file_paths)
        self.fps_dict = {}
        for file_path in self.file_paths:
            _, self.video_name, _ = get_fn_ext(filepath=file_path)
            _, px_per_mm, fps = self.read_video_info(video_name=self.video_name)
            self.fps_dict[self.video_name] = fps
            self.data_df = read_df(file_path=file_path, file_type=self.file_type)
            bp_df = self.data_df[[x for x in self.bp_headers if not x.endswith("_p") and not TAIL_END in x.lower()]]
            p_df = self.data_df[[x for x in self.bp_headers if x.endswith("_p") and not TAIL_END in x.lower()]]
            bp_arr = bp_df.values.reshape(len(bp_df), int(len(bp_df.columns) / 2), 2).astype(np.int64)
            p_arr = p_df.values.reshape(len(p_df), len(p_df.columns), 1)
            if self.threshold > 0.0:
                bp_arr = GeometryMixin.filter_low_p_bps_for_shapes(x=bp_arr, p=p_arr, threshold=self.threshold).reshape(bp_arr.shape[0], -1, 2)
            self.animal_polygons = GeometryMixin().multiframe_bodyparts_to_polygon(data=bp_arr,
                                                                                   parallel_offset=self.buffer,
                                                                                   pixels_per_mm=px_per_mm,
                                                                                   video_name=self.video_name,
                                                                                   animal_name="Animal",
                                                                                   verbose=self.verbose)
            self.roi_df = pd.DataFrame()
            for geo_name, geo in self.roi_geos[self.video_name].items():
                roi_geo = [geo for x in range(len(self.animal_polygons))]
                pct_overlap = np.array(GeometryMixin().multiframe_compute_pct_shape_overlap(shape_1=self.animal_polygons,
                                                                                            shape_2=roi_geo,
                                                                                            denominator="shape_1",
                                                                                            verbose=self.verbose,
                                                                                            animal_names=geo_name,
                                                                                            video_name=self.video_name))
                frames_in_roi = np.zeros(pct_overlap.shape)
                frames_in_roi[np.argwhere(pct_overlap >= self.animal_area)] = 1
                self.roi_df[geo_name] = frames_in_roi

            self.video_results = FeatureExtractionSupplemental.spontaneous_alternations(data=self.roi_df, arm_names=self.arm_names, center_name=self.center_name)
            self.results[self.video_name] = self.video_results

    def save(self):
        print("Running alternation analysis...")
        results_df = pd.DataFrame(columns=["VIDEO NAME","ALTERNATION RATE","ALTERNATION COUNT","ERROR COUNT","SAME ARM RETURN ERRORS","ALTERNATE ARM RETURN ERRORS"])
        save_path = os.path.join(self.logs_path, f"spontaneous_alternation_{self.datetime}.csv")
        for video_name, d in self.results.items():
            results_df.loc[len(results_df)] = [video_name, d["pct_alternation"], d["alternation_cnt"], d["error_cnt"], d["same_arm_returns_cnt"], d["alternate_arm_returns_cnt"]]
        results_df.set_index("VIDEO NAME").to_csv(save_path)
        stdout_success(msg=f"Spontaneous alternation data for {len(list(self.results.keys()))} video(s) saved at {save_path}")
        if self.detailed_data:
            save_dir = os.path.join(self.logs_path, f"detailed_spontaneous_alternation_data_{self.datetime}")
            sliced_keys = ["same_arm_returns_dict","alternate_arm_returns_dict","alternations_dict"]
            replace_keys = {"same_arm_returns_dict": "same arm return","alternate_arm_returns_dict": "alternate arm return","alternations_dict": "alternations",}
            os.makedirs(save_dir)
            for video_name, d in self.results.items():
                details_path = os.path.join(save_dir, f"{video_name}_details.csv")
                arm_entry_sequence_path = os.path.join(save_dir, f"{video_name}_arm_entry_sequence.csv")
                sliced_data = {k: v for k, v in d.items() if k in sliced_keys}
                sliced_data = {replace_keys.get(key, key): value for key, value in sliced_data.items()}
                row_idx = [(o, i) for o, i in sliced_data.items() for i in i.keys()]
                values = [v for i in sliced_data.values() for v in i.values()]
                values = [["" if len(sublist) == 0 else ", ".join(map(str, sublist))] for sublist in values]
                multi_index = pd.MultiIndex.from_tuples(row_idx, names=["Behavior", "Arm"])
                df = pd.DataFrame(values, index=multi_index, columns=["Frames"])
                d['arm_entry_sequence']['ENTER TIME (s)'] = d['arm_entry_sequence']['Start_frame'] / self.fps_dict[video_name]
                d['arm_entry_sequence']['EXIT TIME (s)'] = d['arm_entry_sequence']['End_frame'] / self.fps_dict[video_name]
                arm_entry_sequence_df = pd.DataFrame(data=d['arm_entry_sequence'].values, columns=['ARM_ENTRY_SEQUENCE', 'ENTRY FRAME', 'EXIT FRAME', 'ENTER TIME (s)', 'EXIT TIME (s)'])
                df.to_csv(details_path)
                arm_entry_sequence_df.to_csv(arm_entry_sequence_path)

            stdout_success(msg=f"Detailed spontaneous alternation data for {len(list(self.results.keys()))} video(s) saved at {save_dir}")

# config_path = '/Users/simon/Desktop/envs/simba/troubleshooting/spontenous_alternation/project_folder/project_config.ini'
# arm_names = ['A', 'B', 'C']
# center_name = 'Center'
# threshold = 0.0
# animal_area = 60
# buffer = 2
# verbose = True
#
# x = SpontaneousAlternationCalculator(config_path=config_path,
#                                      arm_names=arm_names,
#                                      center_name=center_name,
#                                      threshold=threshold,
#                                      detailed_data=True,
#                                      animal_area=animal_area,
#                                      buffer=buffer,
#                                      verbose=verbose)
# x.run()
# x.save()



import multiprocessing
import os
import platform
import threading
from tkinter import *
from typing import Union

from simba.data_processors.spontaneous_alternation_calculator import \
    SpontaneousAlternationCalculator
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.spontaneous_alternation_plotter import \
    SpontaneousAlternationsPlotter
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        FileSelect)
from simba.utils.enums import Formats, Keys, Links, Options
from simba.utils.errors import (AnimalNumberError, CountError,
                                InvalidInputError, NoFilesFoundError,
                                NoROIDataError)
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    find_video_of_file, get_fn_ext)


class SpontaneousAlternationPopUp(ConfigReader, PopUpMixin):
    """
    Pop-up window for setting spontaneous alternation parameters and running spontaneous alternation analysis and visualizations.

    :example:
    >>> _ = SpontaneousAlternationPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/spontenous_alternation/project_folder/project_config.ini')
    """

    def __init__(self, config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        if not os.path.isfile(self.roi_coordinates_path):
            raise NoROIDataError(
                msg=f"Spontaneous alternation requires user-defined ROIs. No ROI data found at {self.roi_coordinates_path}",
                source=self.__class__.__name__,
            )
        if self.animal_cnt != 1:
            raise AnimalNumberError(
                msg=f"Spontaneous alternation 1 animal project, found {self.animal_cnt}.",
                source=self.__class__.__name__,
            )
        self.read_roi_data()
        if len(self.shape_names) < 4:
            raise CountError(
                msg=f"Spontaneous alternation requires at least {4} defined ROIs. Got {len(self.shape_names)}.",
                source=self.__class__.__name__,
            )
        if len(self.outlier_corrected_paths) == 0:
            raise NoFilesFoundError(
                msg=f"No data found in {self.outlier_corrected_dir} directory.",
                source=self.__class__.__name__,
            )
        video_files = find_files_of_filetypes_in_directory(
            directory=self.video_dir, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value
        )
        self.video_names = [get_fn_ext(x)[1] for x in video_files]
        PopUpMixin.__init__(self, title="SPONTANEOUS ALTERNATION CALCULATOR", icon='t')
        self.config_path = config_path
        self.arm_dict = self.create_dropdown_frame(
            main_frm=self.main_frm,
            drop_down_titles=[
                "MAZE ARM 1:",
                "MAZE ARM 2:",
                "MAZE ARM 3:",
                "MAZE CENTER:",
            ],
            drop_down_options=self.shape_names,
            frm_title="ARM DEFINITIONS",
        )
        self.animal_settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="ANIMAL SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ROI_FEATURES_PLOT.value,
        )
        self.threshold_dropdown = DropDownMenu(
            self.animal_settings_frm,
            "POSE-ESTIMATION THRESHOLD: ",
            list(range(0, 105, 5)),
            labelwidth=35,
        )
        self.animal_area_dropdown = DropDownMenu(
            self.animal_settings_frm,
            "ANIMAL AREA (%): ",
            list(range(51, 101, 1)),
            labelwidth=35,
        )
        self.buffer_dropdown = DropDownMenu(
            self.animal_settings_frm,
            "ANIMAL BUFFER (MM): ",
            list(range(0, 105, 5)),
            labelwidth=35,
        )
        self.detailed_data_dropdown = DropDownMenu(
            self.animal_settings_frm,
            "SAVE DETAILED DATA: ",
            ["True", "False"],
            labelwidth=35,
        )
        self.verbose_dropdown = DropDownMenu(
            self.animal_settings_frm, "VERBOSE: ", ["True", "False"], labelwidth=35
        )

        self.threshold_dropdown.setChoices(0)
        self.animal_area_dropdown.setChoices(80)
        self.buffer_dropdown.setChoices(0)
        self.detailed_data_dropdown.setChoices("False")
        self.verbose_dropdown.setChoices("False")

        self.animal_settings_frm.grid(row=2, column=0, sticky=NW)
        self.threshold_dropdown.grid(row=0, column=0, sticky=NW)
        self.animal_area_dropdown.grid(row=1, column=0, sticky=NW)
        self.buffer_dropdown.grid(row=2, column=0, sticky=NW)
        self.detailed_data_dropdown.grid(row=3, column=0, sticky=NW)
        self.verbose_dropdown.grid(row=4, column=0, sticky=NW)
        # self.create_run_frm(run_function=self.run_analysis, title="RUN ANALYSIS")

        self.run_analysis_frm = LabelFrame(
            self.main_frm,
            text="RUN ANALYSIS",
            font=Formats.FONT_HEADER.value,
            fg="black",
        )
        self.run_analysis_btn = Button(
            self.run_analysis_frm,
            text="RUN ANALYSIS",
            fg="blue",
            font=Formats.FONT_REGULAR.value,
            command=lambda: threading.Thread(target=self.run_analysis()).start(),
        )

        self.run_analysis_frm.grid(row=5, column=0, sticky=NW)
        self.run_analysis_btn.grid(row=0, column=0, sticky=NW)

        self.run_visualization_frm = LabelFrame(
            self.main_frm,
            text="RUN VISUALIZATION",
            font=Formats.FONT_HEADER.value,
            fg="black",
        )
        self.video_run_btn = Button(
            self.run_visualization_frm,
            text="CREATE VIDEO",
            fg="blue",
            font=Formats.FONT_REGULAR.value,
            command=lambda: self.run_visualization(),
        )
        self.single_video_dropdown = DropDownMenu(
            self.run_visualization_frm,
            "Video:",
            self.video_names,
            "12",
            com=lambda x: self.__update_single_video_file_path(filename=x),
        )
        self.select_video_file_select = FileSelect(
            self.run_visualization_frm,
            "",
            lblwidth="1",
            file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
            dropdown=self.single_video_dropdown,
        )
        self.single_video_dropdown.setChoices(self.video_names[0])
        self.select_video_file_select.filePath.set(self.video_names[0])
        self.run_visualization_frm.grid(row=6, column=0, sticky=NW)
        self.video_run_btn.grid(row=0, column=0, sticky=NW)
        self.single_video_dropdown.grid(row=0, column=1, sticky=NW)
        self.select_video_file_select.grid(row=0, column=2, sticky=NW)
        self.select_video_file_select.grid(row=0, column=3, sticky=NW)
        self.main_frm.mainloop()

    def __update_single_video_file_path(self, filename: str):
        self.select_video_file_select.filePath.set(filename)

    def __checks(self):
        self.arm_names = [
            self.arm_dict["MAZE ARM 1:"].getChoices(),
            self.arm_dict["MAZE ARM 2:"].getChoices(),
            self.arm_dict["MAZE ARM 3:"].getChoices(),
        ]
        self.center_name = self.arm_dict["MAZE CENTER:"].getChoices()
        if self.center_name in self.arm_names:
            raise InvalidInputError(
                msg=f"One ROI has been defined both as an ARM, and as the CENTER: {self.center_name}",
                source=self.__class__.__name__,
            )
        if len(list(set(self.arm_names))) != len(self.arm_names):
            raise InvalidInputError(
                msg=f"Each arm has to be unique but got {self.arm_names}",
                source=self.__class__.__name__,
            )
        self.animal_area = int(self.animal_area_dropdown.getChoices())
        self.threshold = float(self.threshold_dropdown.getChoices()) / 100
        self.buffer = int(self.buffer_dropdown.getChoices()) + 1
        if self.detailed_data_dropdown.getChoices() == "True":
            self.detailed_data = True
        else:
            self.detailed_data = False
        if self.verbose_dropdown.getChoices() == "True":
            self.verbose = True
        else:
            self.verbose = False

    def run_analysis(self):
        self.__checks()
        calculator = SpontaneousAlternationCalculator(
            config_path=self.config_path,
            arm_names=self.arm_names,
            center_name=self.center_name,
            animal_area=self.animal_area,
            threshold=self.threshold,
            buffer=self.buffer,
            verbose=self.verbose,
            detailed_data=self.detailed_data,
        )
        # calculator.run()

        threading.Thread(target=calculator.run()).start()
        calculator.save()

    def run_visualization(self):
        self.__checks()
        file_name = self.single_video_dropdown.getChoices()
        data_path = os.path.join(
            self.outlier_corrected_dir, f"{file_name}.{self.file_type}"
        )
        _ = find_video_of_file(
            video_dir=self.video_dir, filename=file_name, raise_error=True
        )
        plotter = SpontaneousAlternationsPlotter(
            config_path=self.config_path,
            arm_names=self.arm_names,
            center_name=self.center_name,
            buffer=self.buffer,
            verbose=self.verbose,
            data_path=data_path,
        )
        plotter.run()


#_ = SpontaneousAlternationPopUp(config_path=r"C:\troubleshooting\spontenous_alternation\project_folder\project_config.ini")



from tkinter import *

import numpy as np

from simba.data_processors.cuda.utils import _is_cuda_available
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.yolo_pose_visualizer import YOLOPoseVisualizer
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FileSelect,
                                        FolderSelect, SimbaButton,
                                        SimBADropDown)
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists)
from simba.utils.enums import Options, PackageNames
from simba.utils.errors import (NoDataError, SimBAGPUError,
                                SimBAPAckageVersionError)
from simba.utils.printing import stdout_warning
from simba.utils.read_write import (find_core_cnt,
                                    find_files_of_filetypes_in_directory,
                                    get_pkg_version, get_video_meta_data,
                                    str_2_bool)
from simba.utils.warnings import MissingFileWarning

MAX_TRACKS_OPTIONS = ['None', 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
CORE_CNT_OPTIONS = list(range(1, find_core_cnt()[0]))
THRESHOLD_OPTIONS = np.arange(0.1, 1.1, 0.1).astype(np.float32)
SIZE_OPTIONS = list(range(1, 21, 1))
SIZE_OPTIONS.insert(0, 'AUTO')

class YoloPoseVisualizerPopUp(PopUpMixin):

    def __init__(self):
        gpu_available, gpus = _is_cuda_available()
        if not gpu_available:
            raise SimBAGPUError(msg=f'Cannot train YOLO pose-estimation model. No NVIDA GPUs detected on machine', source=self.__class__.__name__)
        ultralytics_version = get_pkg_version(pkg=PackageNames.ULTRALYTICS.value)
        if ultralytics_version is None:
            raise SimBAPAckageVersionError(msg=f'Cannot train YOLO pose-estimation model: Could not find ultralytics package in python environment',  source=self.__class__.__name__)

        PopUpMixin.__init__(self, title="PLOT YOLO POSE ESTIMATION RESULTS", icon='ultralytics_2')
        settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings')
        self.save_dir = FolderSelect(settings_frm, folderDescription="SAVE DIRECTORY:", lblwidth=35, entry_width=45)
        self.core_cnt_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=CORE_CNT_OPTIONS, label="CPU CORE COUNT:", label_width=35, dropdown_width=40, value=int(max(CORE_CNT_OPTIONS) / 2))
        self.bbox_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="SHOW BOUNDING BOXES:",  label_width=35, dropdown_width=40, value='FALSE')
        self.verbose_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="VERBOSE:",  label_width=35, dropdown_width=40, value='TRUE')
        self.threshold_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=THRESHOLD_OPTIONS, label="THRESHOLD:",  label_width=35, dropdown_width=40, value=0.5)
        self.thickness_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=SIZE_OPTIONS, label="LINE THICKNESS:",  label_width=35, dropdown_width=40, value='AUTO')
        self.circle_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=SIZE_OPTIONS, label="CIRCLE SIZE:", label_width=35, dropdown_width=40, value='AUTO')

        settings_frm.grid(row=0, column=0, sticky=NW)
        self.save_dir.grid(row=0, column=0, sticky=NW)
        self.core_cnt_dropdown.grid(row=1, column=0, sticky=NW)
        self.bbox_dropdown.grid(row=2, column=0, sticky=NW)
        self.verbose_dropdown.grid(row=3, column=0, sticky=NW)
        self.threshold_dropdown.grid(row=4, column=0, sticky=NW)
        self.thickness_dropdown.grid(row=5, column=0, sticky=NW)
        self.circle_dropdown.grid(row=6, column=0, sticky=NW)

        single_video_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="PLOT SINGLE VIDEO", icon_name='video')
        self.data_path = FileSelect(parent=single_video_frm, fileDescription='DATA PATH (CSV):', lblwidth=35,  entry_width=45, file_types=[("YOLO CSV RESULT", ".csv")])
        self.video_path = FileSelect(parent=single_video_frm, fileDescription='VIDEO PATH:', lblwidth=35,  entry_width=45, file_types=[("YOLO CSV RESULT", Options.ALL_VIDEO_FORMAT_OPTIONS.value)])
        single_video_btn = SimbaButton(parent=single_video_frm, txt='CREATE SINGLE VIDEO', img='rocket', cmd=self.run, cmd_kwargs={'multiple': False})

        single_video_frm.grid(row=1, column=0, sticky=NW)
        self.data_path.grid(row=0, column=0, sticky=NW)
        self.video_path.grid(row=1, column=0, sticky=NW)
        single_video_btn.grid(row=2, column=0, sticky=NW)

        multiple_video_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="PLOT MULTIPLE VIDEOS", icon_name='stack')
        self.data_dir_path = FolderSelect(parent=multiple_video_frm, folderDescription='DATA DIRECTORY:', lblwidth=35,  entry_width=45)
        self.video_dir_path = FolderSelect(parent=multiple_video_frm, folderDescription='VIDEO DIRECTORY:', lblwidth=35,  entry_width=45)
        multiple_video_btn = SimbaButton(parent=multiple_video_frm, txt='CREATE MULTIPLE VIDEOS', img='rocket', cmd=self.run, cmd_kwargs={'multiple': True})

        multiple_video_frm.grid(row=2, column=0, sticky=NW)
        self.data_dir_path.grid(row=0, column=0, sticky=NW)
        self.video_dir_path.grid(row=1, column=0, sticky=NW)
        multiple_video_btn.grid(row=2, column=0, sticky=NW)

        self.main_frm.mainloop()

    def run(self, multiple: bool):

        save_dir = self.save_dir.folder_path
        core_cnt = int(self.core_cnt_dropdown.get_value())
        bbox = str_2_bool(self.bbox_dropdown.get_value())
        verbose = str_2_bool(self.verbose_dropdown.get_value())
        threshold = float(self.threshold_dropdown.get_value())
        thickness = None if self.threshold_dropdown.get_value() == 'AUTO' else float(self.threshold_dropdown.get_value())
        circle_size = None if self.circle_dropdown.get_value() == 'AUTO' else float(self.circle_dropdown.get_value())
        if not multiple:
            data_path = self.data_path.file_path
            video_path = self.video_path.file_path
            check_file_exist_and_readable(file_path=data_path, raise_error=True)
            _ = get_video_meta_data(video_path=video_path)
            plotter = YOLOPoseVisualizer(data_path=data_path, video_path=video_path, save_dir=save_dir, core_cnt=core_cnt, threshold=threshold, thickness=thickness, circle_size=circle_size, verbose=verbose, bbox=bbox)
            plotter.run()
        else:
            data_dir = self.data_dir_path.folder_path
            video_dir = self.video_dir_path.folder_path
            check_if_dir_exists(in_dir=data_dir)
            check_if_dir_exists(in_dir=video_dir)
            data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=['.csv'], raise_error=True, as_dict=True)
            video_paths = find_files_of_filetypes_in_directory(directory=video_dir, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value, raise_error=True, as_dict=True)

            missing_videos = list([x for x in video_paths.keys() if x not in data_paths.keys()])
            missing_data_paths = list([x for x in data_paths.keys() if x not in video_paths.keys()])
            if len(missing_videos) > 0:
                MissingFileWarning(msg=f'Data files are missing video files in the {video_dir} directory: {missing_videos}', source=self.__class__.__name__)
            if len(missing_data_paths) > 0:
                MissingFileWarning(msg=f'Video files are missing data files in the {data_dir} directory: {missing_data_paths}', source=self.__class__.__name__)

            data_paths = {k:v for k, v in data_paths.items() if k in video_paths.keys()}
            if len(list(data_paths.keys())) == 0:
                raise NoDataError(msg=f'No data file in the {data_dir} directory has a representative video file in the {video_dir} directory', source=self.__class__.__name__)
            video_cnt = len(list(data_paths.keys()))
            for cnt, (name, data_path) in enumerate(data_paths.items()):
                if name in video_paths.keys():
                    video_path = video_paths[name]
                    print(f'Plotting YOLO results for video {name} (video {cnt+1}/{video_cnt})')
                    plotter = YOLOPoseVisualizer(data_path=data_path, video_path=video_path, save_dir=save_dir, core_cnt=core_cnt, threshold=threshold, thickness=thickness, circle_size=circle_size, verbose=verbose, bbox=bbox)
                    plotter.run()
                else:
                    stdout_warning(msg=f'Skipping video {name} (no video exist in {video_dir})...')



#YoloPlotSingleVideoPopUp()







__author__ = "Simon Nilsson"

from tkinter import *

import numpy as np

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.third_party_label_appenders.transform.dlc_to_labelme import \
    DLC2Labelme
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FolderSelect,
                                        SimBADropDown)
from simba.utils.checks import check_if_dir_exists
from simba.utils.read_write import str_2_bool


class DLC2LabelmePopUp(PopUpMixin):

    """
    GUI for converting DeepLabCut (DLC) annotations to YOLO keypoint format.

    :example:
    >>> DLC2LabelmePopUp()
    """

    def __init__(self):
        PopUpMixin.__init__(self, title="DLC MULTI-ANIMAL H5 PREDICTIONS TO YOLO", icon='dlc_1')
        settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings')
        self.dlc_dir = FolderSelect(settings_frm, folderDescription="DLC ANNOTATION DIRECTORY:", lblwidth=35, entry_width=40, initialdir=r'D:\troubleshooting\dlc_h5_multianimal_to_yolo\data')
        self.save_dir = FolderSelect(settings_frm, folderDescription="SAVE DIRECTORY:", lblwidth=35, entry_width=40, initialdir='D:\troubleshooting\dlc_h5_multianimal_to_yolo\yolo')
        self.grey_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="GREYSCALE: ", label_width=35, dropdown_width=40, value='FALSE')
        self.clahe_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="CLAHE: ", label_width=35, dropdown_width=40, value='FALSE')
        self.verbose_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="VERBOSE: ", label_width=35, dropdown_width=40, value='TRUE')


        settings_frm.grid(row=0, column=0, sticky=NW)
        self.dlc_dir .grid(row=0, column=0, sticky=NW)
        self.save_dir.grid(row=1, column=0, sticky=NW)
        self.grey_dropdown.grid(row=2, column=0, sticky=NW)
        self.clahe_dropdown.grid(row=3, column=0, sticky=NW)
        self.verbose_dropdown.grid(row=4, column=0, sticky=NW)

        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()


    def run(self):
        dlc_dir = self.dlc_dir.folder_path
        save_dir = self.save_dir.folder_path

        grey = str_2_bool(self.grey_dropdown.get_value())
        clahe = str_2_bool(self.clahe_dropdown.get_value())
        verbose = str_2_bool(self.verbose_dropdown.get_value())
        check_if_dir_exists(in_dir=dlc_dir, source=f'{self.__class__.__name__} DLC ANNOTATION DIRECTORY', raise_error=True)
        check_if_dir_exists(in_dir=save_dir, source=f'{self.__class__.__name__} SAVE DIRECTORY', raise_error=True)

        runner = DLC2Labelme(dlc_dir=dlc_dir, save_dir=save_dir, verbose=verbose, greyscale=grey, clahe=clahe)
        runner.run()


#DLC2LabelmePopUp()


__author__ = "Simon Nilsson"

import itertools
import os
from typing import Any, Dict, List, Union

import cv2
import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.roi_tools.ROI_feature_analyzer import ROIFeatureCreator
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_keys_exist_in_dict, check_int,
                                check_valid_array, check_valid_dataframe,
                                check_valid_lst,
                                check_video_and_data_frm_count_align)
from simba.utils.data import slice_roi_dict_for_video
from simba.utils.enums import Formats, Keys, TextOptions
from simba.utils.errors import (BodypartColumnNotFoundError, NoFilesFoundError,
                                ROICoordinatesNotFoundError)
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, get_video_meta_data, read_df
from simba.utils.warnings import DuplicateNamesWarning

ROI_CENTERS = "roi_centers"
ROI_EAR_TAGS = "roi_ear_tags"
DIRECTIONALITY = "directionality"
DIRECTIONALITY_STYLE = "directionality_style"
BORDER_COLOR = "border_color"
POSE = "pose_estimation"
ANIMAL_NAMES = "animal_names"

STYLE_KEYS = [ROI_CENTERS,
              ROI_EAR_TAGS,
              DIRECTIONALITY,
              BORDER_COLOR,
              POSE,
              DIRECTIONALITY_STYLE,
              ANIMAL_NAMES]


class ROIfeatureVisualizer(ConfigReader):
    """
    Visualizing features that depend on the relationships between the location of the animals and user-defined
    ROIs. E.g., distances to centroids of ROIs, if animals are directing towards ROIs, and if animals are within ROIs.

    .. note::
       For improved run-time, see :meth:`simba.ROI_feature_visualizer_mp.ROIfeatureVisualizerMultiprocess` for multiprocess class.
       `Tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#part-5-visualizing-roi-features>`__.

    .. image:: _static/img/roi_visualize.png
       :width: 400
       :align: center

    .. image:: _static/img/ROIfeatureVisualizer_1.png
       :width: 700
       :align: center

    .. image:: _static/img/ROIfeatureVisualizer_2.png
       :width: 700
       :align: center

    :param Union[str, os.PathLike] config_path: Path to SimBA project config file in Configparser format
    :param Union[str, os.PathLike] video_path: Path to video file to overlay ROI features on.
    :param List[str] body_parts: List of body-parts to use as proxy for animal location(s).
    :param Dict[str, Any] style_attr: User-defined styles (sizes, colors etc.)

    :example:
    >>> style_attr = {'roi_centers': True, 'roi_ear_tags': True, 'directionality': True, 'directionality_style': 'funnel', 'border_color': (0, 0, 0), 'pose_estimation': True, 'animal_names': True}
    >>> test = ROIfeatureVisualizer(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/project_config.ini', video_path='/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/videos/2022-06-20_NOB_DOT_4.mp4', style_attr=style_attr, body_parts=['Nose'])
    >>> test.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 video_path: Union[str, os.PathLike],
                 body_parts: List[str],
                 style_attr: Dict[str, Any]):

        check_file_exist_and_readable(file_path=config_path)
        check_file_exist_and_readable(file_path=video_path)
        check_if_keys_exist_in_dict(data=style_attr, key=STYLE_KEYS, name=f"{self.__class__.__name__} style_attr")
        _, self.video_name, _ = get_fn_ext(video_path)
        ConfigReader.__init__(self, config_path=config_path)
        if not os.path.isfile(self.roi_coordinates_path):
            raise ROICoordinatesNotFoundError(expected_file_path=self.roi_coordinates_path)
        self.read_roi_data()
        self.roi_dict, shape_names = slice_roi_dict_for_video(data=self.roi_dict, video_name=self.video_name)
        self.data_path = os.path.join(self.outlier_corrected_dir, f"{self.video_name}.{self.file_type}")
        if not os.path.isfile(self.data_path):
            raise NoFilesFoundError(msg=f"SIMBA ERROR: Could not find the file at path {self.data_path}. Make sure the data file exist to create ROI visualizations", source=self.__class__.__name__)
        if not os.path.exists(self.roi_features_save_dir):
            os.makedirs(self.roi_features_save_dir)
        self.save_path = os.path.join(self.roi_features_save_dir, f"{self.video_name}.mp4")
        check_valid_lst(data=body_parts, source=f"{self.__class__.__name__} body-parts", valid_dtypes=(str,), min_len=1)
        for bp in body_parts:
            if bp not in self.body_parts_lst:
                raise BodypartColumnNotFoundError(msg=f"The body-part {bp} is not a valid body-part in the SimBA project. Options: {self.body_parts_lst}", source=self.__class__.__name__)
        self.roi_feature_creator = ROIFeatureCreator(config_path=config_path, body_parts=body_parts, append_data=False, data_path=self.data_path)
        self.roi_feature_creator.run()
        self.bp_lk = self.roi_feature_creator.bp_lk
        self.animal_bp_names = [f"{v[0]} {v[1]}" for v in self.bp_lk.values()]
        self.animal_names = [v[0] for v in self.bp_lk.values()]
        self.video_meta_data = get_video_meta_data(video_path, fps_as_int=False)
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        self.cap = cv2.VideoCapture(video_path)
        check_video_and_data_frm_count_align(video=video_path, data=self.data_path, name=video_path, raise_error=False)
        self.style_attr = style_attr
        self.direct_viable = self.roi_feature_creator.roi_directing_viable
        self.data_df = read_df(file_path=self.data_path, file_type=self.file_type).reset_index(drop=True)
        self.shape_dicts = self.__create_shape_dicts()
        self.directing_df = self.roi_feature_creator.dr

    def __calc_text_locs(self):
        add_spacer = TextOptions.FIRST_LINE_SPACING.value
        self.loc_dict = {}
        txt_strs = []
        for animal_cnt, animal_bp_name in enumerate(self.animal_bp_names):
            for shape in self.shape_names:
                txt_strs.append(animal_bp_name + ' ' + shape + ' center distance')
        longest_text_str = str(max(txt_strs, key=len))
        self.font_size, self.x_scaler, self.y_scaler = PlottingMixin().get_optimal_font_scales(text=longest_text_str, accepted_px_width=int(self.video_meta_data['width'] / 2), accepted_px_height=int(self.video_meta_data['height'] / 15), text_thickness=3)
        print(self.x_scaler, self.y_scaler)
        self.circle_size = PlottingMixin().get_optimal_circle_size(frame_size=(int(self.video_meta_data['width']), int(self.video_meta_data['height'])), circle_frame_ratio=100)
        for animal_cnt, animal_data in self.bp_lk.items():
            animal, animal_bp, _ = animal_data
            animal_name = f"{animal} {animal_bp}"
            self.loc_dict[animal_name] = {}
            self.loc_dict[animal] = {}
            for shape in self.shape_names:
                self.loc_dict[animal_name][shape] = {}
                self.loc_dict[animal_name][shape]["in_zone_text"] = f"{shape} {animal_name} in zone"
                self.loc_dict[animal_name][shape]["distance_text"] = f"{shape} {animal_name} distance"
                self.loc_dict[animal_name][shape]["in_zone_text_loc"] = (int(self.video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value), (self.video_meta_data["height"] - (self.video_meta_data["height"] + 10) + self.y_scaler * add_spacer))
                self.loc_dict[animal_name][shape]["in_zone_data_loc"] = (int(self.video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value + self.x_scaler), (self.video_meta_data["height"] - (self.video_meta_data["height"] + TextOptions.BORDER_BUFFER_Y.value) + self.y_scaler * add_spacer))
                add_spacer += 1
                self.loc_dict[animal_name][shape]["distance_text_loc"] = (int(self.video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value), (self.video_meta_data["height"] - (self.video_meta_data["height"] + 10) + self.y_scaler * add_spacer))
                self.loc_dict[animal_name][shape]["distance_data_loc"] = (int(self.video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value + self.x_scaler), (self.video_meta_data["height"] - (self.video_meta_data["height"] + 10) + self.y_scaler * add_spacer))
                add_spacer += 1
                if self.direct_viable and self.style_attr[DIRECTIONALITY]:
                    self.loc_dict[animal][shape] = {}
                    self.loc_dict[animal][shape]["directing_text"] = f"{shape} {animal} facing"
                    self.loc_dict[animal][shape]["directing_text_loc"] = (int(self.video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value), (self.video_meta_data["height"] - (self.video_meta_data["height"] + 10) + self.y_scaler * add_spacer))
                    self.loc_dict[animal][shape]["directing_data_loc"] = (int(self.video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value + self.x_scaler), (self.video_meta_data["height"] - (self.video_meta_data["height"] + 10)+ self.y_scaler * add_spacer))
                    add_spacer += 1

    def __create_shape_dicts(self):
        shape_dicts = {}
        for shape, df in self.roi_dict.items():
            if not df["Name"].is_unique:
                df = df.drop_duplicates(subset=["Name"], keep="first")
                DuplicateNamesWarning(msg=f'Some of your ROIs with the same shape ({shape}) has the same names for video {self.video_name}. E.g., you have two rectangles named "My rectangle". SimBA prefers ROI shapes with unique names. SimBA will keep one of the unique shape names and drop the rest.', source=self.__class__.__name__)
            d = df.set_index("Name").to_dict(orient="index")
            shape_dicts = {**shape_dicts, **d}
        return shape_dicts

    def __insert_texts(self, shape_df):
        for cnt, animal_data in self.bp_lk.items():
            animal, animal_bp, _ = animal_data
            animal_name = f"{animal} {animal_bp}"
            for _, shape in shape_df.iterrows():
                shape_name, shape_color = shape["Name"], shape["Color BGR"]
                cv2.putText(self.img_w_border, self.loc_dict[animal_name][shape_name]["in_zone_text"], self.loc_dict[animal_name][shape_name]["in_zone_text_loc"], self.font, self.font_size, shape_color, TextOptions.TEXT_THICKNESS.value)
                cv2.putText(self.img_w_border, self.loc_dict[animal_name][shape_name]["distance_text"], self.loc_dict[animal_name][shape_name]["distance_text_loc"], self.font, self.font_size, shape_color, TextOptions.TEXT_THICKNESS.value)
                if self.direct_viable:
                    cv2.putText(self.img_w_border, self.loc_dict[animal][shape_name]["directing_text"], self.loc_dict[animal][shape_name]["directing_text_loc"], self.font, self.font_size, shape_color, TextOptions.TEXT_THICKNESS.value)

    def run(self):
        self.frame_cnt = 0
        while self.cap.isOpened():
            ret, self.img = self.cap.read()
            if ret:
                self.img_w_border = cv2.copyMakeBorder(self.img, 0, 0, 0, self.video_meta_data["width"], borderType=cv2.BORDER_CONSTANT, value=self.style_attr[BORDER_COLOR])
                if self.frame_cnt == 0:
                    self.img_w_border_h, self.img_w_border_w = (self.img_w_border.shape[0], self.img_w_border.shape[1])
                    self.__calc_text_locs()
                    self.writer = cv2.VideoWriter(self.save_path, self.fourcc, self.video_meta_data["fps"], (self.img_w_border_w, self.img_w_border_h))
                self.__insert_texts(self.roi_dict[Keys.ROI_RECTANGLES.value])
                self.__insert_texts(self.roi_dict[Keys.ROI_CIRCLES.value])
                self.__insert_texts(self.roi_dict[Keys.ROI_POLYGONS.value])
                if self.style_attr[POSE]:
                    for animal_name, bp_data in self.animal_bp_dict.items():
                        for bp_cnt, bp in enumerate(zip(bp_data["X_bps"], bp_data["Y_bps"])):
                            bp_cords = self.data_df.loc[self.frame_cnt, list(bp)].values.astype(np.int64)
                            cv2.circle(self.img_w_border, (bp_cords[0], bp_cords[1]), self.circle_size, self.animal_bp_dict[animal_name]["colors"][bp_cnt], -1)
                if self.style_attr[ANIMAL_NAMES]:
                    for animal_name, bp_data in self.animal_bp_dict.items():
                        headers = [bp_data["X_bps"][-1], bp_data["Y_bps"][-1]]
                        bp_cords = self.data_df.loc[self.frame_cnt, headers].values.astype(np.int64)
                        cv2.putText(self.img_w_border, animal_name, (bp_cords[0], bp_cords[1]), self.font, self.font_size, self.animal_bp_dict[animal_name]["colors"][0], TextOptions.TEXT_THICKNESS.value)

                self.img_w_border = PlottingMixin.roi_dict_onto_img(img=self.img_w_border, roi_dict=self.roi_dict, circle_size=self.circle_size, show_tags=self.style_attr[ROI_EAR_TAGS], show_center=self.style_attr[ROI_CENTERS])

                for animal_name, shape_name in itertools.product(self.animal_bp_names, self.shape_names):
                    in_zone_col_name = f"{shape_name} {animal_name} in zone"
                    distance_col_name = f"{shape_name} {animal_name} distance"
                    in_zone_value = str(bool(self.roi_feature_creator.out_df.loc[self.frame_cnt, in_zone_col_name]))
                    distance_value = round(self.roi_feature_creator.out_df.loc[self.frame_cnt, distance_col_name], 2)
                    cv2.putText(self.img_w_border, in_zone_value, self.loc_dict[animal_name][shape_name]["in_zone_data_loc"], self.font, self.font_size, self.shape_dicts[shape_name]["Color BGR"], TextOptions.TEXT_THICKNESS.value)
                    cv2.putText(self.img_w_border,  str(distance_value),  self.loc_dict[animal_name][shape_name]["distance_data_loc"],  self.font,  self.font_size,  self.shape_dicts[shape_name]["Color BGR"],  TextOptions.TEXT_THICKNESS.value)
                if self.direct_viable and self.style_attr[DIRECTIONALITY]:
                    for animal_name, shape_name in itertools.product(self.animal_names, self.shape_names):
                        facing_col_name = f"{shape_name} {animal_name} facing"
                        facing_value = self.roi_feature_creator.out_df.loc[self.frame_cnt, facing_col_name]
                        cv2.putText(self.img_w_border, str(bool(facing_value)), self.loc_dict[animal_name][shape_name]["directing_data_loc"], self.font, self.font_size, self.shape_dicts[shape_name]["Color BGR"], TextOptions.TEXT_THICKNESS.value)
                        if facing_value:
                            self.img_w_border = PlottingMixin.insert_directing_line(directing_df=self.directing_df,
                                                                                    img=self.img_w_border,
                                                                                    shape_name=shape_name,
                                                                                    animal_name=animal_name,
                                                                                    frame_id=self.frame_cnt,
                                                                                    color=self.shape_dicts[shape_name]["Color BGR"],
                                                                                    thickness=self.shape_dicts[shape_name]["Thickness"],
                                                                                    style=self.style_attr[DIRECTIONALITY_STYLE])
                self.frame_cnt += 1
                self.writer.write(np.uint8(self.img_w_border))
                print(f"Frame: {self.frame_cnt} / {self.video_meta_data['frame_count']}. Video: {self.video_name} ...")
            else:
                break

        self.timer.stop_timer()
        self.cap.release()
        self.writer.release()
        stdout_success(
            msg=f"Feature video {self.video_name} saved in {self.save_path} directory ...",
            elapsed_time=self.timer.elapsed_time_str,
        )


# style_attr = {'roi_centers': True, 'roi_ear_tags': True, 'directionality': True, 'directionality_style': 'lines', 'border_color': (0, 0, 0), 'pose_estimation': True, 'animal_names': True}
# test = ROIfeatureVisualizer(config_path=r"C:\troubleshooting\roi_duplicates\project_folder\project_config.ini",
#                             video_path=r"C:\troubleshooting\roi_duplicates\project_folder\videos\2021-12-21_15-03-57_CO_Trimmed.mp4",
#                             style_attr=style_attr,
#                             body_parts=['Butt/Proximal Tail']) #'Butt/Proximal Tail'
# test.run()




# style_attr = {'roi_centers': True, 'roi_ear_tags': True, 'directionality': True, 'directionality_style': 'lines', 'border_color': (0, 0, 0), 'pose_estimation': True, 'animal_names': True}
# test = ROIfeatureVisualizer(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/project_config.ini',
#                             video_path='/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/videos/2022-06-20_NOB_DOT_4.mp4',
#                             style_attr=style_attr,
#                             body_parts=['Nose'])
# test.run()


# style_attr = {'roi_centers': True, 'roi_ear_tags': True, 'directionality': True, 'directionality_style': 'funnel', 'border_color': (0, 128, 0), 'pose_estimation': True, 'animal_names': True}
# test = ROIfeatureVisualizer(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                             video_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/videos/Together_1.avi',
#                             style_attr=style_attr,
#                             body_parts=['Nose_1', 'Nose_2'])
# test.run()


# style_attr = {'ROI_centers': True, 'ROI_ear_tags': True, 'Directionality': True, 'Directionality_style': 'Line', 'Border_color': (0, 128, 0), 'Pose_estimation': True}
# test = ROIfeatureVisualizer(config_path='/Users/simon/Desktop/envs/simba_dev/tests/test_data/mouse_open_field/project_folder/project_config.ini', video_name='Video1.mp4', style_attr=style_attr)
# test.run()


# test = ROIfeatureVisualizer(config_path='/Users/simon/Desktop/train_model_project/project_folder/project_config.ini', video_name='Together_1.avi')
# test.run()
# test.save_new_features_files()



__author__ = "Simon Nilsson"

import os
from typing import Any, Dict, Union

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.unsupervised.enums import Clustering, Unsupervised
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_keys_exist_in_dict, check_instance)
from simba.utils.printing import stdout_success
from simba.utils.read_write import read_pickle

METHOD = "method"
PEARSON = "pearson"
KENDALL = "kendall"
PALETTE = "palette"
SHAP = "shap"
PLOTS = "plots"
CREATE = "create"
SPEARMAN = "spearman"
CORRELATIONS = "correlations"


class EmbeddingCorrelationCalculator(UnsupervisedMixin, ConfigReader):
    """
    Class for correlating dimensionality reduction features with original features for explainability purposes.

    .. image:: _static/img/EmbeddingCorrelationCalculator.png
       :width: 800
       :align: center

    :param str config_path: path to SimBA configparser.ConfigParser project_config.ini
    :param str data_path: path to pickle holding unsupervised results in ``data_map.yaml`` format.
    :param dict settings: dict holding which statistical tests to use and how to create plots.

    :Example:
    >>> settings = {'correlation_methods': ['pearson', 'kendall', 'spearman'], 'plots': {'create': True, 'correlations': 'pearson', 'palette': 'jet'}}
    >>> calculator = EmbeddingCorrelationCalculator(config_path='unsupervised/project_folder/project_config.ini', data_path='unsupervised/cluster_models/quizzical_rhodes.pickle', settings=settings)
    >>> calculator.run()
    """

    def __init__(
        self,
        data_path: Union[str, os.PathLike],
        config_path: Union[str, os.PathLike],
        settings: Dict[str, Any],
    ):

        check_file_exist_and_readable(file_path=config_path)
        check_instance(
            source=f"{self.__class__.__name__} settings",
            instance=settings,
            accepted_types=(dict,),
        )
        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        check_file_exist_and_readable(file_path=data_path)
        self.settings, self.data_path = settings, data_path
        check_if_keys_exist_in_dict(
            data=settings,
            key=[CORRELATIONS, PLOTS],
            name=f"{self.__class__.__name__} settings",
        )
        self.data = read_pickle(data_path=self.data_path)
        check_if_keys_exist_in_dict(
            data=self.data,
            key=[Unsupervised.METHODS.value, Unsupervised.DR_MODEL.value],
            name=self.data_path,
        )
        self.save_path = os.path.join(
            self.logs_path,
            f"embedding_correlations_{self.data[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}_{self.datetime}.csv",
        )

    def run(self):
        print("Calculating embedding correlations...")
        self.x_df = self.data[Unsupervised.METHODS.value][
            Unsupervised.SCALED_DATA.value
        ]
        self.y_df = pd.DataFrame(
            self.data[Unsupervised.DR_MODEL.value][Unsupervised.MODEL.value].embedding_,
            columns=["X", "Y"],
            index=self.x_df.index,
        )
        results = pd.DataFrame()
        for correlation_method in self.settings[CORRELATIONS]:
            results[f"{correlation_method}_Y"] = self.x_df.corrwith(
                self.y_df["Y"], method=correlation_method
            )
            results[f"{correlation_method}_X"] = self.x_df.corrwith(
                self.y_df["X"], method=correlation_method
            )
        results.to_csv(self.save_path)
        self.timer.stop_timer()
        stdout_success(
            msg=f"Embedding correlations saved in {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )

        if self.settings[PLOTS][CREATE]:
            print("Creating embedding correlation plots...")
            df = pd.concat([self.x_df, self.y_df], axis=1)
            save_dir = os.path.join(
                self.logs_path,
                f"embedding_correlation_plots_{self.data[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}_{self.datetime}",
            )
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            for feature_cnt, feature_name in enumerate(
                self.data[Unsupervised.METHODS.value][Unsupervised.FEATURE_NAMES.value]
            ):
                save_path = os.path.join(save_dir, f"{feature_name}.png")
                _ = PlottingMixin.continuous_scatter(
                    data=df,
                    columns=["X", "Y", feature_name],
                    palette=self.settings[PLOTS][PALETTE],
                    title=feature_name,
                    save_path=save_path,
                    show_box=False,
                )
                print(
                    f"Saving image {str(feature_cnt+1)}/{str(len(df.columns))} ({feature_name})"
                )

        self.timer.stop_timer()
        stdout_success(
            msg=f"Embedding correlation calculations complete",
            elapsed_time=self.timer.elapsed_time_str,
        )


# settings = {'correlations': ['pearson', 'kendall', 'spearman'], 'plots': {'create': True, 'correlations': 'pearson', 'palette': 'jet'}}
# calculator = EmbeddingCorrelationCalculator(config_path='/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini',
#                                             data_path='/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/clusters/beautiful_beaver.pickle',
#                                             settings=settings)
# calculator.run()
